from django.contrib import admin
from django.template.loader import render_to_string
from django.urls import reverse
from edc_model.admin import HistoricalModelAdminMixin
from edc_model_admin.history import SimpleHistoryAdmin
from edc_utils.date import to_local

from ....admin_site import edc_pharmacy_history_admin
from ....models import DispenseItem
from ...model_admin_mixin import ModelAdminMixin


@admin.register(DispenseItem.history.model, site=edc_pharmacy_history_admin)
class DispenseItemHistoryAdmin(ModelAdminMixin, HistoricalModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Dispense items (History)"
    change_form_title = "Pharmacy: Dispense item (History)"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20

    list_display = (
        "identifier",
        "subject",
        "history_type",
        "formatted_history_date",
        "code",
        "dispensed_by",
        "dispense_changelist",
        "revision",
        "history_id",
    )

    search_fields = (
        "id",
        "dispense__id",
        "code",
        "dispense__rx__subject_identifier",
    )

    readonly_fields = (
        "dispense_item_identifier",
        "dispense_item_datetime",
        "code",
    )

    @admin.display(description="Dispense #", ordering="-dispense_item_identifier")
    def identifier(self, obj):
        return obj.dispense_item_identifier

    @admin.display(
        description="Subject #",
        ordering="dispense__rx__registered_subject__subject_identifier",
    )
    def subject(self, obj):
        return obj.dispense.rx.registered_subject.subject_identifier

    @admin.display(
        description="Location",
        ordering="dispense__location",
    )
    def location(self, obj):
        return obj.dispense.location

    @admin.display(
        description="dispensed by",
        ordering="dispense__dispensed_by",
    )
    def dispensed_by(self, obj):
        return obj.dispense.dispensed_by

    @admin.display(description="item date", ordering="dispense_item_datetime")
    def dispense_date(self, obj):
        return to_local(obj.dispense_item_datetime).date()

    @admin.display(description="DISPENSE #")
    def dispense_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_dispense_changelist")
        url = f"{url}?q={obj.dispense.id}"
        context = dict(url=url, label=obj.dispense.dispense_identifier, title="Go to dispense")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(description="Stock #")
    def stock_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stockproxy_changelist")
        url = f"{url}?q={obj.stock.code}"
        context = dict(url=url, label=obj.stock.code, title="Go to stock")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    def get_view_only_site_ids_for_user(self, request) -> list[int]:
        return [s.id for s in request.user.userprofile.sites.all() if s.id != request.site.id]



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps

from .mailing_list_manager import MailingListManager
from .site_notifications import site_notifications
from .utils import get_email_enabled

if TYPE_CHECKING:
    from requests.models import Response

    from .models import Notification


def update_mailing_lists_in_m2m(
    sender=None,
    userprofile=None,
    pk_set=None,
    subscribe=None,
    unsubscribe=None,
    verbose=None,
    email_enabled=None,
) -> Response:
    """Update mailing list if email is enabled.

    See post-migrate signal:
        'manage_mailists_on_userprofile_m2m_changed'

    m2m_model = m2m model class for 'email_notifications' or
    'sms_notifications'.
    """
    response = None
    email_enabled = email_enabled or get_email_enabled()
    if (
        email_enabled
        and site_notifications.loaded
        and userprofile.email_notifications.through == sender
    ):
        notification_model_cls: type[Notification] = django_apps.get_model(
            "edc_notification.Notification"
        )
        for notification_obj in notification_model_cls.objects.filter(
            pk__in=list(pk_set), enabled=True
        ):
            notification_cls = site_notifications.get(notification_obj.name)
            notification = notification_cls()
            manager = MailingListManager(
                address=notification.email_to[0],
                display_name=notification.display_name,
                name=notification.name,
            )
            # response = manager.create(verbose=verbose)
            if subscribe:
                response = manager.subscribe(userprofile.user, verbose=verbose)
            elif unsubscribe:
                response = manager.unsubscribe(userprofile.user, verbose=verbose)
    return response



from uuid import uuid4

from clinicedc_constants import CLOSED, RESOLVED
from django.core.exceptions import ObjectDoesNotExist
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models.deletion import PROTECT
from django.template.loader import render_to_string
from django.urls.base import reverse
from django.utils import timezone

from edc_action_item.models.action_model_mixin import ActionModelMixin
from edc_dashboard.url_names import InvalidDashboardUrlName, url_names
from edc_model.models import BaseUuidModel
from edc_sites.managers import CurrentSiteManager
from edc_sites.model_mixins import SiteModelMixin
from edc_visit_tracking.utils import get_related_visit_model_cls

from ..action_items import DATA_QUERY_ACTION
from ..constants import AUTO_RESOLVED
from .data_dictionary import DataDictionary
from .model_mixins import DataQueryModelMixin
from .query_subject import QuerySubject
from .query_visit_schedule import QueryVisitSchedule
from .requisition_panel import RequisitionPanel


class DataQuery(DataQueryModelMixin, ActionModelMixin, SiteModelMixin, BaseUuidModel):
    action_name = DATA_QUERY_ACTION

    subject_identifier = models.CharField(max_length=50, default="", editable=False)

    registered_subject = models.ForeignKey(
        QuerySubject,
        verbose_name="Subject Identifier",
        on_delete=PROTECT,
        null=True,
        blank=False,
    )

    visit_schedule = models.ForeignKey(
        QueryVisitSchedule,
        verbose_name="Visit",
        on_delete=PROTECT,
        null=True,
        blank=True,
    )

    visit_code_sequence = models.IntegerField(
        verbose_name="Visit code sequence",
        validators=[MinValueValidator(0), MaxValueValidator(25)],
        null=True,
        blank=True,
    )

    timepoint = models.DecimalField(null=True, decimal_places=1, max_digits=6)

    data_dictionaries = models.ManyToManyField(
        DataDictionary,
        verbose_name="CRF question(s)",
        blank=True,
        help_text="select all that apply",
    )

    requisition_panel = models.ForeignKey(
        RequisitionPanel,
        verbose_name="Are responses linked to a requisition? If so, which",
        related_name="+",
        on_delete=PROTECT,
        null=True,
        blank=True,
        help_text="Requisition will be expected on day of visit.",
    )

    missed_visit = models.BooleanField(
        verbose_name="Visit reported as missed",
        default=False,
        help_text="If visit/timepoint was missed, data is not expected",
    )

    rule_generated = models.BooleanField(
        default=False, help_text="This query was auto-generated by a query rule."
    )

    rule_reference = models.CharField(
        verbose_name="Query rule reference", max_length=150, default=uuid4
    )

    objects = models.Manager()

    on_site = CurrentSiteManager()

    def __str__(self):
        return f"{self.action_identifier[-9:]}, {self.status}"

    def save(self, *args, **kwargs):
        self.subject_identifier = self.registered_subject.subject_identifier
        self.try_to_resolve_auto_generated_query()
        super().save(*args, **kwargs)

    def try_to_resolve_auto_generated_query(self):
        if (
            not self.dm_resolved
            and self.rule_generated
            and self.site_response_status == RESOLVED
            and self.site_response_text
            and self.site_response_text.strip() == AUTO_RESOLVED
        ):
            self.status = CLOSED
            self.resolved_datetime = timezone.now()
            self.dm_user = self.sender

    def form_and_numbers_to_string(self):
        ret = []
        model_verbose_names = [
            o.model_verbose_name for o in self.data_dictionaries.all().order_by("model")
        ]
        model_verbose_names = list(set(model_verbose_names))
        for model_verbose_name in model_verbose_names:
            numbers = [
                str(o.number)
                for o in self.data_dictionaries.filter(
                    model_verbose_name=model_verbose_name
                ).order_by("number")
            ]
            numbers = ", ".join(numbers)
            ret.append((model_verbose_name, numbers))
        return ret

    def get_action_item_display_name(self):
        pass

    def get_action_item_reason(self):
        try:
            url = url_names.get("subject_dashboard_url")
        except InvalidDashboardUrlName:
            visit_href = "#"
        else:
            try:
                visit = get_related_visit_model_cls().objects.get(
                    subject_identifier=self.registered_subject.subject_identifier,
                    visit_schedule_name=self.visit_schedule.visit_schedule_name,
                    schedule_name=self.visit_schedule.schedule_name,
                    visit_code=self.visit_schedule.visit_code,
                    visit_code_sequence=self.visit_code_sequence,
                )
            except ObjectDoesNotExist:
                visit_href = "#"
            except AttributeError as e:
                if (
                    "visit_schedule_name" not in str(e)
                    and "schedule_name" not in str(e)
                    and "visit_code" not in str(e)
                ):
                    raise
                visit_href = "#"
            else:
                visit_href = reverse(
                    url,
                    kwargs=dict(
                        appointment=str(visit.appointment.id),
                        subject_identifier=self.registered_subject.subject_identifier,
                    ),
                )

        template_name = "edc_data_manager/columns/query_text.html"
        context = dict(
            form_and_numbers=self.form_and_numbers_to_string(),
            query_priority=self.query_priority,
            query_priority_display=self.get_query_priority_display(),
            query_text=self.query_text,
            questions=self.data_dictionaries.all().order_by("model", "number"),
            report_datetime=self.report_datetime,
            requisition_panel=self.requisition_panel,
            resolved_datetime=self.resolved_datetime,
            site_resolved_datetime=self.site_resolved_datetime,
            site_response_text=self.site_response_text,
            site_response_status=self.get_site_response_status_display(),
            status=self.status,
            dm_user=self.dm_user,
            title=self.title,
            visit_schedule=self.visit_schedule,
            visit_href=visit_href,
        )
        return render_to_string(template_name, context=context)

    @property
    def model_names(self):
        model_names = list(set([dd.model for dd in self.data_dictionaries.all()]))
        return "|".join(model_names)

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Data Query"
        verbose_name_plural = "Data Queries"
        unique_together = ("registered_subject", "rule_reference", "visit_schedule")
        indexes = (
            models.Index(
                fields=[
                    "subject_identifier",
                    "action_identifier",
                    "title",
                    "rule_reference",
                    "registered_subject",
                    "visit_schedule",
                ]
            ),
        )



# Generated by Django 5.2.3 on 2025-06-27 13:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "edc_export",
            "0019_alter_datarequest_site_alter_datarequesthistory_site_and_more",
        ),
        ("sites", "0002_alter_domain_unique"),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name="datarequesthistory",
            name="edc_export__exporte_ba8050_idx",
        ),
        migrations.RemoveIndex(
            model_name="exportreceipt",
            name="edc_export__timesta_6a2406_idx",
        ),
        migrations.RemoveIndex(
            model_name="filehistory",
            name="edc_export__sent_da_01059e_idx",
        ),
        migrations.RemoveIndex(
            model_name="uploadexportreceiptfile",
            name="edc_export__created_f54b3c_idx",
        ),
        migrations.AddIndex(
            model_name="datarequesthistory",
            index=models.Index(
                fields=["exported_datetime"], name="edc_export__exporte_d57cf7_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="exportreceipt",
            index=models.Index(fields=["timestamp"], name="edc_export__timesta_a855d0_idx"),
        ),
        migrations.AddIndex(
            model_name="filehistory",
            index=models.Index(
                fields=["sent_datetime"], name="edc_export__sent_da_10748e_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="uploadexportreceiptfile",
            index=models.Index(fields=["created"], name="edc_export__created_bd7ff7_idx"),
        ),
    ]



from django.conf import settings


def get_languages_from_settings() -> dict[str, str]:
    """Returns a dictionary of language codes mapped to language names,
    for all languages defined in settings.LANGUAGES.
    """
    try:
        lang_iterator = settings.LANGUAGES.items()
    except AttributeError:
        lang_iterator = settings.LANGUAGES
    return {k: v for k, v in lang_iterator}



from django.db import models
from django.utils import timezone
from sequences import get_next_value

from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_randomization.site_randomizers import site_randomizers
from edc_registration.models import RegisteredSubject

from ...exceptions import AllocationError
from .. import Assignment, Rx
from .stock_request_item import StockRequestItem


class Manager(models.Manager):
    use_in_migrations = True


class Allocation(BaseUuidModel):
    """A model to track stock allocation to a subject referring to a
    stock request.
    """

    allocation_identifier = models.CharField(
        max_length=36,
        unique=True,
        null=True,
        blank=True,
        help_text="A sequential unique identifier set by the EDC",
    )

    allocation_datetime = models.DateTimeField(default=timezone.now)

    registered_subject = models.ForeignKey(
        RegisteredSubject,
        verbose_name="Allocated to",
        on_delete=models.PROTECT,
        null=True,
        blank=False,
    )

    assignment = models.ForeignKey(Assignment, on_delete=models.PROTECT, null=True, blank=True)

    stock_request_item = models.OneToOneField(
        StockRequestItem,
        verbose_name="Requested",
        on_delete=models.PROTECT,
        null=True,
        blank=False,
    )

    allocated_by = models.CharField(max_length=50, default="", blank=True)

    subject_identifier = models.CharField(
        max_length=50, default="", blank=True, editable=False
    )

    code = models.CharField(
        verbose_name="Stock code",
        max_length=15,
        unique=True,
        null=True,
        blank=True,
        help_text="A unique alphanumeric code",
        editable=False,
    )

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return self.allocation_identifier

    def save(self, *args, **kwargs):
        if not self.allocation_identifier:
            self.allocation_identifier = f"{get_next_value(self._meta.label_lower):06d}"
        if not self.stock_request_item:
            raise AllocationError("Stock request item may not be null")
        self.subject_identifier = self.registered_subject.subject_identifier
        # self.code = self.stock.code
        self.assignment = self.get_assignment()
        super().save(*args, **kwargs)

    def get_assignment(self) -> Assignment:
        rx = Rx.objects.get(
            registered_subject=RegisteredSubject.objects.get(id=self.registered_subject.id)
        )
        randomizer = site_randomizers.get(rx.randomizer_name)
        assignment = randomizer.get_assignment(self.registered_subject.subject_identifier)
        return Assignment.objects.get(name=assignment)

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Allocation"
        verbose_name_plural = "Allocations"



from __future__ import annotations

from clinicedc_constants import CHOL, DM, HIV, HTN, NO
from django.db import models

from edc_constants.choices import YES_NO
from edc_dx import get_diagnosis_labels

default_prompts = {
    HIV.lower(): "Has the patient ever tested <U>positive</U> for HIV infection?",
    DM: "Has the patient ever been diagnosed with DIABETES?",
    HTN: "Has the patient ever been diagnosed with HYPERTENSION?",
    CHOL: "Has the patient ever been diagnosed with HIGH CHOLESTEROL?",
}


def baseline_review_model_mixin_factory(prompts: dict[str, str] | None = None):
    prompts = prompts or default_prompts

    class AbstractModel(models.Model):
        class Meta:
            abstract = True

    opts = {}
    for cond, label in get_diagnosis_labels().items():
        opts.update(
            {
                f"{cond}_dx": models.CharField(
                    verbose_name=prompts.get(cond),
                    max_length=15,
                    choices=YES_NO,
                ),
                f"{cond}_dx_at_screening": models.CharField(
                    verbose_name=f"Was a diagnosis of {label.upper()} reported at screening?",
                    max_length=15,
                    choices=YES_NO,
                ),
            }
        )

    opts.update(
        {
            "protocol_incident": models.CharField(
                verbose_name=(
                    "Do any of the above diagnosis differ from those reported at screening?"
                ),
                max_length=15,
                choices=YES_NO,
                default=NO,
            )
        }
    )
    for name, fld_cls in opts.items():
        AbstractModel.add_to_class(name, fld_cls)

    return AbstractModel



# Generated by Django 5.2.10 on 2026-01-24 15:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0106_auto_20260119_2332"),
    ]

    operations = [
        migrations.RenameField(
            model_name="container",
            old_name="qty",
            new_name="qty_suggested",
        ),
        migrations.RenameField(
            model_name="historicalcontainer",
            old_name="qty",
            new_name="qty_suggested",
        ),
        migrations.AddField(
            model_name="historicalreceiveitem",
            name="container_qty",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="Number of items in this/these container(s). If more than one container is received here, all containers have the same container quantity",
                max_digits=20,
                null=True,
                verbose_name="Container quantity",
            ),
        ),
        migrations.AddField(
            model_name="receiveitem",
            name="container_qty",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="Number of items in this/these container(s). If more than one container is received here, all containers have the same container quantity",
                max_digits=20,
                null=True,
                verbose_name="Container quantity",
            ),
        ),
    ]



from __future__ import annotations

from clinicedc_utils import ConversionNotHandled
from django.core.exceptions import ObjectDoesNotExist

from .molecular_weight_model_cls import molecular_weight_model_cls

__all__ = ["get_mw"]


def get_mw(label):
    try:
        molecular_weight = molecular_weight_model_cls().objects.get(label=label)
    except ObjectDoesNotExist as e:
        raise ConversionNotHandled(
            f"Conversion not handled. Molecular weight not found for {label}."
        ) from e
    else:
        mw = molecular_weight.mw
    return mw



from .crf_metadata import CrfMetadata
from .requisition_metadata import RequisitionMetadata
from .signals import (
    metadata_create_on_post_save,
    metadata_reset_on_post_delete,
    metadata_update_on_post_save,
)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django_audit_fields import audit_fieldset_tuple

if TYPE_CHECKING:
    from edc_lab import RequisitionPanel

panel_conclusion_fieldset: tuple[str, dict] = (
    "Conclusion",
    {"fields": ("results_abnormal", "results_reportable")},
)
panel_summary_fieldset: tuple[str, dict] = (
    "Summary",
    {"fields": ("reportable_summary", "abnormal_summary", "errors")},
)


calculate_egfr_fieldset: tuple[str, dict] = (
    "Calculated eGFR",
    {
        # "classes": ("collapse",),
        "description": "To be calculated (or recalculated) when this form is saved",
        "fields": ["egfr_value", "egfr_units", "egfr_grade"],
    },
)

calculate_egfr_drop_fieldset: tuple[str, dict] = (
    "Calculated eGFR Drop",
    {
        # "classes": ("collapse",),
        "description": "To be calculated (or recalculated) when this form is saved",
        "fields": ["egfr_drop_value", "egfr_drop_units", "egfr_drop_grade"],
    },
)


class BloodResultFieldsetError(Exception):
    pass


class BloodResultFieldset:
    """A class to generate a modeladmin `fieldsets` using the
    lab panel for this `blood result`.
    """

    def __init__(
        self,
        panel: RequisitionPanel,
        title: str = None,
        model_cls: Any = None,
        extra_fieldsets: list[tuple[int, tuple[str, dict]]] | None = None,
        excluded_utest_ids: list[str] = None,
        exclude_units: bool = None,
        exclude_reportable: bool = None,
    ):
        self.panel = panel
        self.title = (title or panel.name).replace("_", " ").title()
        self.model_cls = model_cls
        self.extra_fieldsets = extra_fieldsets
        self.excluded_utest_ids = excluded_utest_ids or []
        self.exclude_units = exclude_units
        self.exclude_reportable = exclude_reportable

    def __repr__(self):
        return f"{self.__class__.__name__}({self.panel})"

    def __str__(self):
        return f"{self.__class__.__name__}({self.panel})"

    @property
    def fieldsets(self):
        fieldsets = [
            (None, {"fields": ("subject_visit", "report_datetime")}),
            (
                "Requisition and Result Date",
                {"fields": ["requisition", "assay_datetime"]},
            ),
        ]
        for utest_id in self.panel.utest_ids:
            if utest_id in self.excluded_utest_ids:
                continue
            if isinstance(utest_id, (tuple,)):
                code, title = utest_id
            else:
                code = utest_id
                title = code.upper()
            fieldsets.append(self.get_panel_item_fieldset(code, title=title))
        if not self.exclude_reportable:
            fieldsets.extend([panel_conclusion_fieldset, panel_summary_fieldset])
        fieldsets.append(audit_fieldset_tuple)
        for pos, fieldset in self.extra_fieldsets or []:
            if pos < 0:
                fieldsets.append(fieldset)
            else:
                fieldsets.insert(pos, fieldset)
        return tuple(fieldsets)

    def get_panel_item_fieldset(self, code, title=None):
        if not title:
            title = code.upper()
        model_fields = [
            f"{code}_value",
            f"{code}_units",
        ]
        if not self.exclude_reportable:
            model_fields.extend(
                [
                    f"{code}_abnormal",
                    f"{code}_reportable",
                ]
            )

        if self.exclude_units:
            model_fields.remove(f"{code}_units")
        if self.model_cls:
            for field in model_fields:
                try:
                    getattr(self.model_cls, field)
                except AttributeError as e:
                    raise BloodResultFieldsetError(f"{e}. See {self}")

        return (
            title,
            {"fields": model_fields},
        )



from django.contrib import admin, messages
from django.db.models import QuerySet
from django.utils.translation import gettext

from ...models import ReceiveItem, Stock


@admin.display(description=f"Delete selected {ReceiveItem._meta.verbose_name_plural}")
def delete_receive_items_action(modeladmin, request, queryset: QuerySet[ReceiveItem]):
    failed_count = 0
    success_count = 0
    for obj in queryset:
        if obj.stock_set.filter(confirmation__isnull=False).exists():
            failed_count += 1
        else:
            Stock.objects.filter(receive_item=obj, confirmation__isnull=True).delete()
            obj.delete()
            success_count += 1
    if success_count > 0:
        messages.add_message(
            request,
            messages.SUCCESS,
            gettext("Successfully deleted %(success_count)s %(verbose_name)s.")
            % dict(success_count=success_count, verbose_name=ReceiveItem._meta.verbose_name),
        )
    if failed_count > 0:
        messages.add_message(
            request,
            messages.ERROR,
            gettext(
                "Unable to deleted %(failed_count)s %(verbose_name)s. "
                "Confirmed stock items exist."
            )
            % dict(failed_count=failed_count, verbose_name=ReceiveItem._meta.verbose_name),
        )



numeric_datatypes = [
    "DecimalField",
    "IntegerField",
    "BigIntegerField",
    "FloatField",
    "PositiveBigIntegerField",
    "PositiveIntegerField",
    "PositiveSmallIntegerField",
    "SmallIntegerField",
]

date_datatypes = ["DateTimeField", "DateField"]

timedelta_datatypes = ["DurationField"]



from .crf import OffstudyCrfModelFormMixin
from .offstudy_modelform_mixins import OffstudyModelFormMixin
from .prn import OffstudyNonCrfModelFormMixin
from .requisition import OffstudyRequisitionModelFormMixin



from django.apps import AppConfig as DjangoAppConfig
from django.core.checks.registry import register

from .system_checks import randomizationlist_check


class AppConfig(DjangoAppConfig):
    name = "edc_randomization"
    verbose_name = "Edc Randomization"
    has_exportable_data = True
    include_in_administration_section = True

    def ready(self):
        register(randomizationlist_check, deploy=True)

    # @property
    # def randomization_list_path(self):
    #     warn(
    #         "Use of settings.RANDOMIZATION_LIST_PATH has been deprecated. "
    #         "See site_randomizers in edc_randomization",
    #         stacklevel=2,
    #     )
    #     return os.path.join(settings.RANDOMIZATION_LIST_PATH)



from clinicedc_utils import (
    ConversionNotHandled,
    EgfrCalculatorError,
    EgfrCkdEpi2009,
    EgfrCkdEpi2021,
    EgfrCockcroftGault,
)
from django import forms

from edc_form_validators import INVALID_ERROR


class EgfrCkdEpiFormValidatorMixin:
    calculator_version = 2009

    @property
    def egfr_calculator_cls(self) -> type[EgfrCkdEpi2009 | EgfrCkdEpi2021]:
        if self.calculator_version == 2009:  # noqa: PLR2004
            cls = EgfrCkdEpi2009
        elif self.calculator_version == 2021:  # noqa: PLR2004
            cls = EgfrCkdEpi2021
        else:
            raise EgfrCalculatorError(
                f"Invalid calculator version. Got {self.calculator_version}."
            )
        return cls

    def validate_egfr(
        self, *, gender: str, age_in_years: int, ethnicity: str | None = None
    ) -> float | None:
        opts = dict(
            gender=gender,
            age_in_years=age_in_years,
            creatinine_value=self.cleaned_data.get("creatinine_value"),
            creatinine_units=self.cleaned_data.get("creatinine_units"),
        )
        if self.egfr_calculator_cls == EgfrCkdEpi2009:
            opts.update(ethnicity=ethnicity)
        try:
            value = self.egfr_calculator_cls(**opts).value
        except (EgfrCalculatorError, ConversionNotHandled) as e:
            raise forms.ValidationError(e) from e
        return value


class EgfrCockcroftGaultFormValidatorMixin:
    def validate_egfr(
        self, *, gender: str, age_in_years: int, weight_in_kgs: float
    ) -> float | None:
        opts = dict(
            gender=gender,
            age_in_years=age_in_years,
            weight=weight_in_kgs,
            creatinine_value=self.cleaned_data.get("creatinine_value"),
            creatinine_units=self.cleaned_data.get("creatinine_units"),
        )
        try:
            value = EgfrCockcroftGault(**opts).value
        except (EgfrCalculatorError, ConversionNotHandled) as e:
            self.raise_validation_error({"__all__": str(e)}, INVALID_ERROR, exc=e)
        return value



from django.apps import AppConfig as DjangoApponfig
from django.core.management import color_style

style = color_style()


class AppConfig(DjangoApponfig):
    name = "edc_model_to_dataframe"
    verbose_name = "Edc model to pandas dataframe"
    include_in_administration_section = False



from django.contrib import admin
from django_audit_fields import ModelAdminAuditFieldsMixin
from django_revision.modeladmin_mixin import ModelAdminRevisionMixin

from edc_model_admin.mixins import (
    ModelAdminHideDeleteButtonOnCondition,
    ModelAdminInstitutionMixin,
    ModelAdminNextUrlRedirectMixin,
    ModelAdminRedirectAllToChangelistMixin,
    ModelAdminRedirectOnDeleteMixin,
    TemplatesModelAdminMixin,
)
from edc_sites.admin import SiteModelAdminMixin

from ..admin_site import edc_export_admin
from ..models import DataRequestHistory


@admin.register(DataRequestHistory, site=edc_export_admin)
class DataRequestHistoryAdmin(
    SiteModelAdminMixin,
    TemplatesModelAdminMixin,
    ModelAdminRedirectOnDeleteMixin,
    ModelAdminRevisionMixin,
    ModelAdminInstitutionMixin,
    ModelAdminNextUrlRedirectMixin,
    ModelAdminAuditFieldsMixin,
    ModelAdminRedirectAllToChangelistMixin,
    ModelAdminHideDeleteButtonOnCondition,
    admin.ModelAdmin,
):
    date_hierarchy = "exported_datetime"

    show_cancel = True
    view_on_site = False
    show_history_label = False

    change_search_field_name = "id"

    fields = (
        "data_request",
        "archive_filename",
        "emailed_to",
        "emailed_datetime",
        "summary",
        "exported_datetime",
    )

    list_display = (
        "data_request",
        "emailed_to",
        "emailed_datetime",
        "exported_datetime",
    )

    list_filter = ("emailed_to", "emailed_datetime", "exported_datetime")

    readonly_fields = (
        "data_request",
        "archive_filename",
        "emailed_to",
        "emailed_datetime",
        "summary",
        "exported_datetime",
    )

    search_fields = ("id", "summary", "archive_filename")

    def hide_delete_button_on_condition(self, request, object_id) -> bool:
        return True


class DataRequestHistoryInline(admin.TabularInline):
    model = DataRequestHistory

    fields = ("archive_filename", "emailed_to", "emailed_datetime", "exported_datetime")

    list_display = ("emailed_to", "exported_datetime", "created")

    readonly_fields = (
        "data_request",
        "archive_filename",
        "emailed_to",
        "emailed_datetime",
        "summary",
        "exported_datetime",
    )

    extra = 0



visit_schedule_fields: tuple[str, ...] = (
    "visit_schedule_name",
    "schedule_name",
    "visit_code",
)

visit_schedule_fieldset_tuple: tuple[str, dict[str, tuple[str, ...]]] = (
    "Visit Schedule",
    {"classes": ("collapse",), "fields": visit_schedule_fields},
)

visit_schedule_only_fields: tuple[str, ...] = (
    "visit_schedule_name",
    "schedule_name",
)

visit_schedule_only_fieldset_tuple: tuple[str, dict[str, tuple[str, ...]]] = (
    "Visit Schedule",
    {"classes": ("collapse",), "fields": visit_schedule_only_fields},
)



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from django.core.exceptions import ObjectDoesNotExist

from ..utils import get_rxrefill_model_cls
from .refill_creator import RefillCreator

if TYPE_CHECKING:
    from ..models import DosageGuideline, Formulation, RxRefill


def create_refill(
    refill_identifier: str,
    subject_identifier: str,
    dosage_guideline: DosageGuideline,
    formulation: Formulation,
    refill_start_datetime: datetime,
    refill_end_datetime: datetime,
    roundup_divisible_by: int | None,
    weight_in_kgs: float | None,
) -> RxRefill:
    """Creates or updated the edc_pharmacy refill for this study medication CRF,
    if not already created.

    Usually called by study medication CRF

    Called from signal.
    """
    rx_refill = None
    creator = None
    try:
        rx_refill = get_rxrefill_model_cls().objects.get(refill_identifier=refill_identifier)
    except ObjectDoesNotExist:
        creator = RefillCreator(
            refill_identifier=refill_identifier,
            dosage_guideline=dosage_guideline,
            formulation=formulation,
            make_active=True,
            refill_start_datetime=refill_start_datetime,
            refill_end_datetime=refill_end_datetime,
            roundup_divisible_by=roundup_divisible_by,
            subject_identifier=subject_identifier,
            weight_in_kgs=weight_in_kgs,
        )
        # adjust_previous
    else:
        rx_refill.dosage_guideline = dosage_guideline
        rx_refill.formulation = formulation
        rx_refill.refill_start_datetime = refill_start_datetime
        rx_refill.refill_end_datetime = refill_end_datetime
        rx_refill.roundup_divisible_by = roundup_divisible_by
        rx_refill.weight_in_kgs = weight_in_kgs
        rx_refill.save()
        rx_refill.refresh_from_db()
    return rx_refill or creator.rx_refill



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-03-21 09:19
from __future__ import unicode_literals

import uuid

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_registration", "0006_auto_20170209_0924")]

    operations = [
        migrations.AlterField(
            model_name="registeredsubject",
            name="subject_identifier_as_pk",
            field=models.UUIDField(default=uuid.uuid4, editable=False),
        )
    ]



# Generated by Django 4.2.1 on 2023-07-07 19:32

from django.db import migrations

import edc_pharmacy.models.prescription.rx_refill
import edc_sites.models


class Migration(migrations.Migration):
    dependencies = [
        (
            "edc_pharmacy",
            "0017_alter_box_options_alter_dispensinghistory_options_and_more",
        ),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="rxrefill",
            managers=[
                ("objects", edc_pharmacy.models.prescription.rx_refill.Manager()),
                ("on_site", edc_sites.models.CurrentSiteManager()),
            ],
        ),
    ]



"""To customize any of the values below,
use settings.LAB_DASHBOARD_URL_NAMES.

These have nothing to do with url_names dictionary
"""

import contextlib

from django.conf import settings

dashboard_urls = dict(
    aliquot_listboard_url="edc_lab_dashboard:aliquot_listboard_url",
    lab_dashboard_home_url="edc_lab_dashboard:home_url",
    # home_url="edc_lab_dashboard:home_url",
    # listboard_url="edc_lab_dashboard:requisition_listboard_url",
    manage_box_listboard_url="edc_lab_dashboard:manage_box_listboard_url",
    manage_manifest_listboard_url="edc_lab_dashboard:manage_manifest_listboard_url",
    manifest_listboard_url="edc_lab_dashboard:manifest_listboard_url",
    pack_listboard_url="edc_lab_dashboard:pack_listboard_url",
    process_listboard_url="edc_lab_dashboard:process_listboard_url",
    receive_listboard_url="edc_lab_dashboard:receive_listboard_url",
    requisition_listboard_url="edc_lab_dashboard:requisition_listboard_url",
    result_listboard_url="edc_lab_dashboard:result_listboard_url",
    verify_box_listboard_url="edc_lab_dashboard:verify_box_listboard_url",
    print_manifest_url="edc_lab_dashboard:print_manifest_url",
    aliquot_form_action_url="edc_lab_dashboard:aliquot_form_action_url",
    manage_box_item_form_action_url="edc_lab_dashboard:manage_box_item_form_action_url",
    manage_manifest_item_form_action_url=(
        "edc_lab_dashboard:manage_manifest_item_form_action_url"
    ),
    manifest_form_action_url="edc_lab_dashboard:manifest_form_action_url",
    pack_form_action_url="edc_lab_dashboard:pack_form_action_url",
    process_form_action_url="edc_lab_dashboard:process_form_action_url",
    receive_form_action_url="edc_lab_dashboard:receive_form_action_url",
    requisition_form_action_url="edc_lab_dashboard:requisition_form_action_url",
    verify_box_item_form_action_url="edc_lab_dashboard:verify_box_item_form_action_url",
)

with contextlib.suppress(AttributeError):
    dashboard_urls.update(**settings.LAB_DASHBOARD_URL_NAMES)



# Generated by Django 5.1.5 on 2025-01-22 03:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_locator", "0037_alter_historicalsubjectlocator_consent_model_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalsubjectlocator",
            name="may_call_work",
            field=models.CharField(
                choices=[("Yes", "Yes"), ("No", "No")],
                max_length=25,
                verbose_name="Has the participant given permission to be contacted <b>at work</b>, by telephone or cell, by study staff for follow-up purposes during the study? ",
            ),
        ),
        migrations.AlterField(
            model_name="subjectlocator",
            name="may_call_work",
            field=models.CharField(
                choices=[("Yes", "Yes"), ("No", "No")],
                max_length=25,
                verbose_name="Has the participant given permission to be contacted <b>at work</b>, by telephone or cell, by study staff for follow-up purposes during the study? ",
            ),
        ),
    ]



from clinicedc_constants import YES
from django.db import models

from edc_constants.choices import YES_NO
from edc_model import models as edc_models
from edc_model.utils import duration_to_date


class NcdInitialReviewModelMixin(models.Model):
    ncd_condition_label: str = None

    med_start_ago = edc_models.DurationYMDField(
        verbose_name=(
            "If the patient is taking medicines for ncd_condition_label, "
            "how long have they been taking these?"
        ),
        null=True,
        blank=True,
    )

    med_start_estimated_date = models.DateField(
        verbose_name="Estimated medication start date",
        null=True,
        editable=False,
    )

    med_start_date_is_estimated = models.CharField(
        verbose_name="Was the medication start date estimated?",
        max_length=15,
        choices=YES_NO,
        default=YES,
        editable=False,
    )

    def save(self, *args, **kwargs):
        if self.med_start_ago:
            self.med_start_estimated_date = duration_to_date(
                self.med_start_ago, self.report_datetime
            )
        super().save(*args, **kwargs)

    class Meta:
        abstract = True



import pandas as pd
from clinicedc_constants import FEMALE, MALE

from ..constants import N_ONLY, N_WITH_COL_PROP, N_WITH_ROW_PROP
from ..row import RowDefinition, RowDefinitions
from ..table import Table


class ArtTable(Table):
    def __init__(self, main_df: pd.DataFrame = None):
        super().__init__(colname=None, main_df=main_df, title="HIV Care")

    @property
    def row_definitions(self) -> RowDefinitions:
        df_tmp = self.main_df.copy()
        row_defs = RowDefinitions(reverse_rows=False)
        row0 = RowDefinition(
            title=self.title,
            label=self.default_sublabel,
            condition=(df_tmp["gender"].notna()),
            columns={FEMALE: (N_ONLY, 2), MALE: (N_ONLY, 2), "All": (N_ONLY, 2)},
            drop=False,
        )
        row_defs.add(row0)
        columns = {
            FEMALE: (N_WITH_COL_PROP, 2),
            MALE: (N_WITH_COL_PROP, 2),
            "All": (N_WITH_ROW_PROP, 2),
        }

        cond_art_stable = (
            (df_tmp["on_rx_stable"] == "Yes")
            & (df_tmp["vl_undetectable"] == "Yes")
            & (df_tmp["art_six_months"] == "Yes")
        )
        row_defs.add(
            RowDefinition(
                label="Stable on ART (6m)",
                condition=cond_art_stable,
                columns=columns,
                drop=True,
            )
        )

        # look for anyone that is not stable by these three values
        condition_partially_stable = (
            (df_tmp["on_rx_stable"] == "Yes")
            | (df_tmp["vl_undetectable"] == "Yes")
            | (df_tmp["art_six_months"] == "Yes")
        )
        row_defs.add(
            RowDefinition(
                label="Other stable (`VL` or `on ART` or `stable 6m`)",
                condition=condition_partially_stable,
                columns=columns,
                drop=True,
            )
        )

        condition_other = (
            (df_tmp["on_rx_stable"] == "No")
            | (df_tmp["vl_undetectable"] == "No")
            | (df_tmp["art_six_months"] == "No")
        )

        row_defs.add(
            RowDefinition(
                label="Other",
                condition=condition_other,
                columns=columns,
                drop=True,
            )
        )

        condition_not_recorded = (
            (df_tmp["on_rx_stable"].isna())
            & (df_tmp["vl_undetectable"].isna())
            & (df_tmp["art_six_months"].isna())
        )
        row_defs.add(
            RowDefinition(
                label="Not recorded",
                condition=condition_not_recorded,
                columns=columns,
                drop=True,
            )
        )
        return row_defs



# Generated by Django 2.2.2 on 2019-08-01 21:28

import _socket
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django.utils.timezone
import django_revision.revision_field
from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="AeClassification",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        db_index=True,
                        help_text="(suggest 40 characters max.)",
                        max_length=250,
                        unique=True,
                        verbose_name="Name",
                    ),
                ),
                (
                    "short_name",
                    models.CharField(
                        db_index=True,
                        help_text="This is the stored value, required",
                        max_length=250,
                        unique=True,
                        verbose_name="Stored value",
                    ),
                ),
                (
                    "display_index",
                    models.IntegerField(
                        db_index=True,
                        default=0,
                        help_text="Index to control display order if not alphabetical, not required",
                        verbose_name="display index",
                    ),
                ),
                (
                    "field_name",
                    models.CharField(
                        blank=True,
                        editable=False,
                        help_text="Not required",
                        max_length=25,
                        null=True,
                    ),
                ),
                (
                    "version",
                    models.CharField(default="1.0", editable=False, max_length=35),
                ),
            ],
            options={"ordering": ["display_index", "name"], "abstract": False},
        ),
        migrations.CreateModel(
            name="SaeReason",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        db_index=True,
                        help_text="(suggest 40 characters max.)",
                        max_length=250,
                        unique=True,
                        verbose_name="Name",
                    ),
                ),
                (
                    "short_name",
                    models.CharField(
                        db_index=True,
                        help_text="This is the stored value, required",
                        max_length=250,
                        unique=True,
                        verbose_name="Stored value",
                    ),
                ),
                (
                    "display_index",
                    models.IntegerField(
                        db_index=True,
                        default=0,
                        help_text="Index to control display order if not alphabetical, not required",
                        verbose_name="display index",
                    ),
                ),
                (
                    "field_name",
                    models.CharField(
                        blank=True,
                        editable=False,
                        help_text="Not required",
                        max_length=25,
                        null=True,
                    ),
                ),
                (
                    "version",
                    models.CharField(default="1.0", editable=False, max_length=35),
                ),
            ],
            options={"ordering": ["display_index", "name"], "abstract": False},
        ),
        migrations.AddIndex(
            model_name="saereason",
            index=models.Index(
                fields=["id", "name", "display_index"], name="edc_adverse_id_88072b_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="aeclassification",
            index=models.Index(
                fields=["id", "name", "display_index"], name="edc_adverse_id_8133d9_idx"
            ),
        ),
    ]



from __future__ import annotations

import getpass
import re
from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING

from django import forms
from django.apps import apps as django_apps
from django.conf import settings
from django.contrib import messages
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.core.management import CommandError
from django.utils import timezone
from django.utils.html import format_html

from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_sites.site import sites as site_sites

from .constants import EXPORT, EXPORT_PII
from .exceptions import ExporterExportFolder
from .files_emailer import FilesEmailer, FilesEmailerError
from .models_to_file import ModelsToFile

if TYPE_CHECKING:
    from django.contrib.auth.base_user import AbstractBaseUser
    from django.contrib.auth.models import User


def get_export_folder() -> Path:
    if path := getattr(settings, "EDC_EXPORT_EXPORT_FOLDER", None):
        return Path(path).expanduser()
    return Path(settings.MEDIA_ROOT) / "data_folder" / "export"


def get_base_dir() -> Path:
    """Returns the base_dir used by, for example,
    shutil.make_archive.

    This is the short protocol name in lower case
    """
    base_dir: str = ResearchProtocolConfig().protocol_lower_name
    if len(base_dir) > 25:
        raise ExporterExportFolder(
            f"Invalid basedir, too long. Using `protocol_lower_name`. Got `{base_dir}`."
        )
    if not re.match(r"^[a-z0-9]+(?:_[a-z0-9]+)*$", base_dir):
        raise ExporterExportFolder(
            "Invalid base_dir, invalid characters. Using `protocol_lower_name`. "
            f"Got `{base_dir}`."
        )
    return Path(base_dir)


def get_upload_folder() -> Path:
    if path := getattr(settings, "EDC_EXPORT_UPLOAD_FOLDER", None):
        return Path(path).expanduser()
    return Path(settings.MEDIA_ROOT) / "data_folder" / "upload"


def get_export_pii_users() -> list[str]:
    return getattr(settings, "EDC_EXPORT_EXPORT_PII_USERS", [])


def raise_if_prohibited_from_export_pii_group(username: str, groups: Iterable) -> None:
    """A user form validation to prevent adding an unlisted
    user to the EXPORT_PII group.

    See also edc_auth's UserForm.
    """
    if EXPORT_PII in [grp.name for grp in groups] and username not in get_export_pii_users():
        raise forms.ValidationError(
            {
                "groups": format_html(
                    "This user is not allowed to export PII data. You may not add "
                    "this user to the <U>{text}</U> group.",
                    text="EXPORT_PII",
                )
            }
        )


def email_files_to_user(request, models_to_file: ModelsToFile) -> None:
    try:
        FilesEmailer(
            path_to_files=models_to_file.tmp_folder,
            user=request.user,
            file_ext=".zip",
            export_filenames=models_to_file.exported_filenames,
        )
    except (FilesEmailerError, ConnectionRefusedError) as e:
        messages.error(request, f"Failed to send files by email. Got '{e}'")
    else:
        messages.success(
            request,
            (
                f"Your data request has been sent to {request.user.email}. "
                "Please check your email."
            ),
        )


def update_data_request_history(request, models_to_file: ModelsToFile):
    summary = [str(x) for x in models_to_file.exported_filenames]
    summary.sort()
    data_request_model_cls = django_apps.get_model("edc_export.datarequest")
    data_request_history_model_cls = django_apps.get_model("edc_export.datarequesthistory")
    data_request = data_request_model_cls.objects.create(
        name=f"Data request {timezone.now().strftime('%Y%m%d%H%M')}",
        models="\n".join(models_to_file.models),
        user_created=request.user.username,
        site=request.site,
    )
    data_request_history_model_cls.objects.create(
        data_request=data_request,
        exported_datetime=timezone.now(),
        summary="\n".join(summary),
        user_created=request.user.username,
        user_modified=request.user.username,
        archive_filename=models_to_file.archive_filename or "",
        emailed_to=models_to_file.emailed_to,
        emailed_datetime=models_to_file.emailed_datetime,
        site=request.site,
    )


def get_export_user() -> User | AbstractBaseUser:
    username = input("Username:")
    passwd = getpass.getpass("Password for " + username + ":")
    try:
        user = get_user_model().objects.get(
            username=username, is_superuser=False, is_active=True
        )
    except ObjectDoesNotExist as e:
        raise CommandError("Invalid username or password.") from e
    if not user.check_password(passwd):
        raise CommandError("Invalid username or password.")
    return user


def validate_user_perms_or_raise(user: User, decrypt: bool | None) -> None:
    if not user.groups.filter(name=EXPORT).exists():
        raise CommandError("You are not authorized to export data.")
    if decrypt and not user.groups.filter(name="EXPORT_PII").exists():
        raise CommandError("You are not authorized to export sensitive data.")


def get_default_models_for_export(trial_prefix: str) -> list[str]:
    apps = [
        f"{trial_prefix}_consent",
        f"{trial_prefix}_lists",
        f"{trial_prefix}_subject",
        f"{trial_prefix}_ae",
        f"{trial_prefix}_prn",
        f"{trial_prefix}_screening",
    ]
    model_names = [
        "edc_appointment.appointment",
        "edc_data_manager.datadictionary",
        "edc_metadata.crfmetadata",
        "edc_metadata.requisitionmetadata",
        "edc_registration.registeredsubject",
        "edc_visit_schedule.subjectschedulehistory",
    ]

    # prepare a list of model names in label lower format
    for app_config in django_apps.get_app_configs():
        if app_config.name.startswith(trial_prefix) and app_config.name in apps:
            model_names.extend(
                [
                    model_cls._meta.label_lower
                    for model_cls in app_config.get_models()
                    if "historical" not in model_cls._meta.label_lower
                    and not model_cls._meta.proxy
                ]
            )
    return model_names


def get_model_names_for_export(
    app_labels: list[str] | None,
    model_names: list[str] | None,
) -> list[str]:
    """Returns a unique list of label_lower"""
    model_names = model_names or []
    if app_labels:
        for app_label in app_labels:
            app_config = django_apps.get_app_config(app_label)
            model_names.extend([cls._meta.label_lower for cls in app_config.get_models()])
    return list(set(model_names))


def get_site_ids_for_export(
    site_ids: list[int] | None,
    countries: list[str] | None,
) -> list[int]:
    """Returns a list of site ids"""
    if countries and site_ids:
        raise CommandError("Invalid. Specify `site_ids` or `countries`, not both.")
    for site_id in site_ids or []:
        try:
            obj = django_apps.get_model("sites.site").objects.get(id=int(site_id))
        except ObjectDoesNotExist as e:
            raise CommandError(f"Invalid site_id. Got `{site_id}`.") from e
        else:
            site_ids.append(obj.id)
    for country in countries or []:
        site_ids.extend(list(site_sites.get_by_country(country)))
    return site_ids



from __future__ import annotations

from uuid import uuid4

from django.contrib import admin, messages
from django.db.models import QuerySet
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import gettext as _

from ...models import Receive, RepackRequest, Stock


@admin.display(description="Confirm repacked and labeled stock")
def confirm_repacked_stock_action(modeladmin, request, queryset: QuerySet[RepackRequest]):
    if queryset.count() > 1 or queryset.count() == 0:
        messages.add_message(
            request,
            messages.ERROR,
            _("Select one and only one item"),
        )
    else:
        return confirm_stock_from_queryset(modeladmin, request, queryset)
    return None


@admin.display(description="Confirm received and labeled stock")
def confirm_received_stock_action(modeladmin, request, queryset: QuerySet[RepackRequest]):
    if queryset.count() > 1 or queryset.count() == 0:
        messages.add_message(
            request,
            messages.ERROR,
            _("Select one and only one item"),
        )
    else:
        return confirm_stock_from_queryset(modeladmin, request, queryset)
    return None


@admin.display(description="Confirm labeled stock")
def confirm_stock_from_queryset(
    modeladmin, request, queryset: QuerySet[RepackRequest | Receive | Stock]
):
    stock_codes = []
    pk = None
    identifier = None
    model_name = None
    word = "confirmed"
    if queryset.count() == 1:
        if queryset.model == RepackRequest:
            stock_qs = Stock.objects.filter(
                repack_request__pk__in=[obj.pk for obj in queryset]
            )
            stock_codes = stock_qs.values_list("code", flat=True)
            pk = queryset.first().pk
            identifier = queryset.first().repack_identifier
            model_name = queryset.model._meta.verbose_name
        elif queryset.model == Receive:
            stock_qs = Stock.objects.filter(
                receive_item__receive__pk__in=[obj.pk for obj in queryset]
            )
            stock_codes = stock_qs.values_list("code", flat=True)
            pk = queryset.first().pk
            identifier = queryset.first().receive_identifier
            model_name = queryset.model._meta.verbose_name
    elif queryset.count() > 1:
        stock_codes = queryset.values_list("code", flat=True)

    if queryset.count() >= 1:
        session_uuid = str(uuid4())
        request.session[session_uuid] = {
            "stock_codes": [str(o) for o in stock_codes],
            "source_pk": pk,
            "source_identifier": identifier,
            "source_label_lower": queryset.model._meta.label_lower,
            "source_model_name": model_name,
            "transaction_word": word,
        }
        url = reverse(
            "edc_pharmacy:confirm_stock_from_queryset_url",
            kwargs={"session_uuid": session_uuid},
        )
        return HttpResponseRedirect(url)
    return None



from __future__ import annotations

from typing import Any

from django.db import models

from ..utils import raise_if_offstudy


class OffstudyNonCrfModelError(Exception):
    pass


class OffstudyNonCrfModelMixin(models.Model):
    """Model mixin for non-CRF, PRN, visit, appt models.

    A mixin for non-CRF models to add the ability to determine
    if the subject is off study as of this non-CRFs report_datetime.

    Requires the fields `subject_identifier`, `report_datetime`.
    """

    def save(self: Any, *args, **kwargs):
        self.raise_if_offstudy()
        super().save(*args, **kwargs)

    def raise_if_offstudy(self) -> None:
        raise_if_offstudy(
            source_obj=self,
            subject_identifier=self.subject_identifier,
            report_datetime=self.report_datetime,
        )

    class Meta:
        abstract = True



from __future__ import annotations

from datetime import datetime

import inflect
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.views.generic.base import TemplateView
from edc_dashboard.view_mixins import EdcViewMixin
from edc_navbar import NavbarViewMixin
from edc_protocol.view_mixins import EdcProtocolViewMixin

from ..constants import CENTRAL_LOCATION
from ..exceptions import StorageBinError
from ..models import Stock, StorageBin, StorageBinItem

p = inflect.engine()


def update_bin(
    storage_bin: StorageBin,
    stock_codes: list[str],
    user_created: str | None = None,
    created: datetime | None = None,
) -> tuple[list[str], list[str]]:
    codes_created = []
    codes_not_created = []
    for code in stock_codes:
        if storage_bin.location.name == CENTRAL_LOCATION:
            opts = dict(
                code=code,
                confirmed=True,
                dispenseitem__isnull=True,
                location=storage_bin.location,
                container=storage_bin.container,
            )
        else:
            opts = dict(
                code=code,
                allocation__isnull=False,
                stocktransferitem__isnull=False,
                stocktransferitem__confirmationatlocationitem__isnull=False,
                dispenseitem__isnull=True,
                stocktransferitem__stock_transfer__to_location=storage_bin.location,
                location=storage_bin.location,
                container=storage_bin.container,
            )
        try:
            stock_obj = Stock.objects.get(**opts)
        except ObjectDoesNotExist:
            stock_obj = None
            codes_not_created.append(code)
        if stock_obj:
            try:
                StorageBinItem.objects.get(stock=stock_obj)
            except ObjectDoesNotExist:
                StorageBinItem.objects.create(
                    stock=stock_obj,
                    storage_bin=storage_bin,
                    user_created=user_created,
                    created=created,
                )
                codes_created.append(code)
            else:
                codes_not_created.append(code)
    return codes_created, codes_not_created


@method_decorator(login_required, name="dispatch")
class AddToStorageBinView(EdcViewMixin, NavbarViewMixin, EdcProtocolViewMixin, TemplateView):
    model_pks: list[str] | None = None
    template_name: str = "edc_pharmacy/stock/add_to_storage_bin.html"
    navbar_name = settings.APP_NAME
    navbar_selected_item = "pharmacy"
    action_word = "Add"

    def get_context_data(self, **kwargs):
        kwargs.update(
            storage_bin=self.storage_bin,
            storage_bin_changelist_url=self.storage_bin_changelist_url,
            items_to_scan_as_range=[],
            SHORT_DATE_FORMAT=settings.SHORT_DATE_FORMAT,
            action_word=self.action_word,
        )
        if self.kwargs.get("items_to_scan"):
            items_to_scan = int(self.kwargs.get("items_to_scan"))
            kwargs.update(items_to_scan_as_range=range(1, items_to_scan + 1))
        return super().get_context_data(**kwargs)

    @property
    def storage_bin(self):
        storage_bin_id = self.kwargs.get("storage_bin")
        try:
            storage_bin = StorageBin.objects.get(id=storage_bin_id)
        except ObjectDoesNotExist:
            storage_bin = None
            messages.add_message(self.request, messages.ERROR, "Invalid storage bin.")
        return storage_bin

    @property
    def storage_bin_changelist_url(self) -> str:
        if self.storage_bin:
            url = reverse("edc_pharmacy_admin:edc_pharmacy_storagebin_changelist")
            return f"{url}?q={self.storage_bin.bin_identifier}"
        return "/"

    def redirect_on_has_duplicates(
        self, stock_codes: list[str], storage_bin: StorageBin
    ) -> HttpResponseRedirect | None:
        if len(stock_codes or []) != len(list(set(stock_codes or []))):
            messages.add_message(
                self.request,
                messages.ERROR,
                "Nothing saved. Duplicate codes detected in list. Please try again.",
            )
            url = reverse(
                "edc_pharmacy:add_to_storage_bin_url",
                kwargs={
                    "stock_request": storage_bin.id,
                },
            )
            return HttpResponseRedirect(url)
        return None

    def redirect_on_stock_already_in_bin(
        self, stock_codes: list[str], storage_bin: StorageBin
    ) -> HttpResponseRedirect | None:
        if (
            stock_codes
            and StorageBinItem.objects.filter(stock__code__in=stock_codes)
            .exclude(storage_bin_id=storage_bin.id)
            .exists()
        ):
            qs = StorageBinItem.objects.filter(stock__code__in=stock_codes)
            messages.add_message(
                self.request,
                messages.ERROR,
                f"Stock already in another bin. See {[obj.stock.code for obj in qs]}.",
            )
            url = reverse(
                "edc_pharmacy:add_to_storage_bin_url",
                kwargs={
                    "storage_bin": storage_bin.id,
                },
            )
            return HttpResponseRedirect(url)
        return None

    def redirect_on_invalid_subject_for_location(
        self, stock_codes: list[str], storage_bin: StorageBin
    ) -> HttpResponseRedirect | None:
        if stock_codes and (
            Stock.objects.filter(code__in=stock_codes)
            .exclude(location=storage_bin.location)
            .exists()
        ):
            qs = Stock.objects.filter(code__in=stock_codes).exclude(
                location=storage_bin.location
            )
            url = reverse("edc_pharmacy_admin:edc_pharmacy_stock_changelist")
            stock_links = [f'<a href="{url}?q={obj.code}">{obj.code}</a>' for obj in qs]
            messages.add_message(
                self.request,
                messages.ERROR,
                format_html("Stock not from this location. See {}", mark_safe(stock_links)),  # noqa: S308
            )
            url = reverse(
                "edc_pharmacy:add_to_storage_bin_url",
                kwargs={
                    "storage_bin": storage_bin.id,
                },
            )
            return HttpResponseRedirect(url)
        return None

    def post(self, request, *args, **kwargs):  # noqa: ARG002
        stock_codes = request.POST.getlist("codes") if request.POST.get("codes") else None
        storage_bin = StorageBin.objects.get(id=kwargs.get("storage_bin"))
        items_to_scan = request.POST.get("items_to_scan") or kwargs.get("items_to_scan")
        if items_to_scan:
            items_to_scan = int(items_to_scan)

        self.redirect_on_has_duplicates(stock_codes, storage_bin)
        self.redirect_on_invalid_subject_for_location(stock_codes, storage_bin)
        self.redirect_on_stock_already_in_bin(stock_codes, storage_bin)
        if items_to_scan and not stock_codes:
            url = reverse(
                "edc_pharmacy:add_to_storage_bin_url",
                kwargs={
                    "storage_bin": storage_bin.id,
                    "items_to_scan": items_to_scan,
                },
            )
            return HttpResponseRedirect(url)
        if items_to_scan and stock_codes:
            try:
                codes_created, codes_not_created = update_bin(
                    storage_bin,
                    stock_codes,
                    user_created=request.user.username,
                    created=timezone.now(),
                )
            except StorageBinError as e:
                messages.add_message(request, messages.ERROR, str(e))
            else:
                messages.add_message(
                    request,
                    messages.SUCCESS,
                    (
                        f"Updated {p.no('stock item', len(codes_created))} to bin. "
                        f"Skipped {len(codes_not_created)}."
                    ),
                )
            return HttpResponseRedirect(self.storage_bin_changelist_url)
        url = reverse(
            "edc_pharmacy:add_to_storage_bin_url",
            kwargs={
                "storage_bin": storage_bin.id,
            },
        )
        return HttpResponseRedirect(url)



# Generated by Django 4.2.7 on 2023-12-05 19:47

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_data_manager", "0031_alter_dataquery_visit_code_sequence_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="datadictionary",
            options={
                "default_permissions": ("view", "export"),
                "verbose_name": "Data Dictionary Item",
                "verbose_name_plural": "Data Dictionary Items",
            },
        ),
        migrations.AlterModelOptions(
            name="dataquery",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Data Query",
                "verbose_name_plural": "Data Queries",
            },
        ),
        migrations.AlterModelOptions(
            name="queryrule",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Query Rule",
                "verbose_name_plural": "Query Rules",
            },
        ),
    ]



from django.db import models
from django.db.models import Index
from django.db.models.deletion import PROTECT
from django.utils import timezone

from edc_model.models import BaseUuidModel
from edc_sites.model_mixins import SiteModelMixin

from .data_request import DataRequest


class DataRequestHistory(SiteModelMixin, BaseUuidModel):
    data_request = models.ForeignKey(DataRequest, on_delete=PROTECT)

    archive_filename = models.CharField(max_length=200, default="")

    emailed_to = models.EmailField(default="")

    emailed_datetime = models.DateTimeField(null=True)

    summary = models.TextField(default="")

    exported_datetime = models.DateTimeField(default=timezone.now)

    class Meta:
        verbose_name = "Data Request History"
        verbose_name_plural = "Data Request History"
        indexes = (Index(fields=["exported_datetime"]),)



from __future__ import annotations

from copy import copy

from django.db.models.constants import LOOKUP_SEP
from django_crypto_fields.utils import get_encrypted_fields

from edc_auth.constants import PII, PII_VIEW


class ModelAdminProtectPiiMixin:
    """Removes encrypted fields from changelist.

    Place first in the MRO.

    Removes encrypted fields and methods returning encrypted
    field values from changelist for users not in PII / PII_VIEW
    groups.

    IMPORTANT: If you declare a method which returns PII to
    the changelist, add the method name to `extra_pii_attrs`.
    """

    extra_pii_attrs: tuple[str] | None = ()

    def get_extra_pii_attrs(self) -> tuple[str | tuple[str, str]]:
        return self.extra_pii_attrs

    def get_encrypted_fields(self) -> tuple[str, ...]:
        encrypted_fields = tuple([f.name for f in get_encrypted_fields(self.model)])
        return tuple({*encrypted_fields, *self.get_extra_pii_attrs()})

    def get_list_display(self, request) -> tuple[str]:
        list_display = super().get_list_display(request)
        if not request.user.groups.filter(name__in=[PII, PII_VIEW]).exists():
            # TODO: search replace from list_display if extra_pii_attr has tuple
            list_display = tuple(
                *{f for f in list_display if f not in self.get_encrypted_fields()}
            )
        return list_display

    def get_list_display_links(self, request, list_display) -> list[str] | None:
        list_display_links = super().get_list_display_links(request, list_display)
        if not request.user.groups.filter(name__in=[PII, PII_VIEW]).exists():
            return None
        return list_display_links

    def get_search_fields(self, request) -> tuple[str]:
        search_fields = super().get_search_fields(request)
        if search_fields:
            search_fields = list(search_fields)
            if not request.user.groups.filter(name__in=[PII, PII_VIEW]).exists():
                for field in search_fields:
                    if field.split(LOOKUP_SEP)[0] in self.get_encrypted_fields():
                        search_fields.remove(field)
            else:
                encrypted_fields = self.get_encrypted_fields()
                for field in copy(search_fields):
                    if field.split(LOOKUP_SEP)[0] in encrypted_fields:
                        try:
                            field.split(LOOKUP_SEP)[1]
                        except IndexError:
                            index = search_fields.index(field)
                            search_fields[index] = f"{field}{LOOKUP_SEP}exact"
            search_fields = tuple(search_fields)
        return search_fields



from django.contrib import admin

from ..get_form_runner_by_src_id import get_form_runner_by_src_id


@admin.action(description="Refresh selected issues")
def issue_refresh(modeladmin, request, queryset):
    for issue_obj in queryset:
        runner = get_form_runner_by_src_id(
            src_id=issue_obj.src_id, model_name=issue_obj.label_lower
        )
        runner.run_one()



from .get_languages_from_settings import get_languages_from_settings


class SiteLanguagesError(Exception):
    pass


def get_languages(language_codes: list[str], site_id: int) -> dict[str, str]:
    defined_languages = get_languages_from_settings()
    if language_codes:
        if unknown_language_codes := [c for c in language_codes if c not in defined_languages]:
            raise SiteLanguagesError(
                "Unknown language code(s) associated with site. Language code must be "
                "defined in settings.LANGUAGES. "
                f"Expected one of {list(defined_languages.keys())}. "
                f"Got {unknown_language_codes} for site `{site_id}`."
            )
        languages = {code: defined_languages[code] for code in language_codes}
    else:
        languages = defined_languages
    return languages



# Generated by Django 6.0.1 on 2026-01-29 19:07

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0133_auto_20260126_0216"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="storagebinproxy",
            options={
                "verbose_name": "Central Storage bin",
                "verbose_name_plural": "Central Storage bins",
            },
        ),
    ]



from clinicedc_constants import QUESTION_RETIRED

def append_question_retired_choice(choices) -> tuple[tuple[str, str], ...]:
    choices = list(choices)
    choices.append((QUESTION_RETIRED, QUESTION_RETIRED))
    return tuple(choices)



from django.contrib import admin
from edc_model.admin import HistoricalModelAdminMixin
from edc_model_admin.history import SimpleHistoryAdmin

from ....admin_site import edc_pharmacy_history_admin
from ....models import Stock
from ....utils import format_qty
from ...model_admin_mixin import ModelAdminMixin


@admin.register(Stock.history.model, site=edc_pharmacy_history_admin)
class StockHistoryAdmin(ModelAdminMixin, HistoricalModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Stock (History)"
    change_form_title = "Pharmacy: Stock (History)"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20

    list_display = (
        "code",
        "history_type",
        "formatted_history_date",
        "subject_identifier",
        "location",
        "formatted_confirmed",
        "formatted_allocation",
        "formatted_in_transit",
        "formatted_confirmed_at_location",
        "formatted_dispensed",
        "qty",
        "formatted_unit_qty_in",
        "formatted_unit_qty_out",
        "unit_qty_in_out",
        "revision",
        "history_id",
    )

    search_fields = ("code",)

    @admin.display(description="C", ordering="confirmed", boolean=True)
    def formatted_confirmed(self, obj):
        return obj.confirmed

    @admin.display(description="A", boolean=True)
    def formatted_allocation(self, obj):
        return bool(obj.subject_identifier)

    @admin.display(description="T", ordering="in_transit", boolean=True)
    def formatted_in_transit(self, obj):
        return obj.in_transit

    @admin.display(description="CL", ordering="confirmed_at_location", boolean=True)
    def formatted_confirmed_at_location(self, obj):
        return obj.confirmed_at_location

    @admin.display(description="D", ordering="dispensed", boolean=True)
    def formatted_dispensed(self, obj):
        return obj.dispensed

    @admin.display(description="QTY", ordering="qty")
    def qty(self, obj):
        return format_qty(obj.qty_in - obj.qty_out, obj.container)

    @admin.display(description="IN", ordering="unit_qty_in")
    def formatted_unit_qty_in(self, obj):
        return format_qty(obj.unit_qty_in, obj.container)

    @admin.display(description="OUT", ordering="unit_qty_out")
    def formatted_unit_qty_out(self, obj):
        return format_qty(obj.unit_qty_out, obj.container)

    @admin.display(description="BAL", ordering="unit_qty_out")
    def unit_qty_in_out(self, obj):
        return format_qty(obj.unit_qty_in - obj.unit_qty_out, obj.container)



# Generated by Django 3.2.13 on 2022-09-09 23:14

from django.db import migrations


def delete_old_permissions(apps, schema_editor):
    model_cls = apps.get_model("auth.permission")
    model_cls._default_manager.filter(
        content_type__app_label="edc_dashboard", content_type__model="dashboard"
    ).delete()


class Migration(migrations.Migration):
    dependencies = [
        ("edc_dashboard", "0001_initial"),
    ]

    operations = [migrations.RunPython(delete_old_permissions)]



# Generated by Django 5.1.2 on 2024-11-19 16:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0046_remove_historicalrx_slug_remove_rx_slug"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalstockrequest",
            name="cancel",
            field=models.CharField(
                default=None,
                max_length=6,
                null=True,
                verbose_name="To cancel this request, type the word 'CANCEL' here and save the form:",
            ),
        ),
        migrations.AddField(
            model_name="stockrequest",
            name="cancel",
            field=models.CharField(
                default=None,
                max_length=6,
                null=True,
                verbose_name="To cancel this request, type the word 'CANCEL' here and save the form:",
            ),
        ),
        migrations.AlterField(
            model_name="historicalstockrequest",
            name="status",
            field=models.CharField(
                choices=[
                    ("open", "Open"),
                    ("closed", "Closed"),
                    ("cancelled", "Cancelled"),
                ],
                default="open",
                max_length=25,
            ),
        ),
        migrations.AlterField(
            model_name="stockrequest",
            name="status",
            field=models.CharField(
                choices=[
                    ("open", "Open"),
                    ("closed", "Closed"),
                    ("cancelled", "Cancelled"),
                ],
                default="open",
                max_length=25,
            ),
        ),
    ]



from .future_date_list_filter import FutureDateListFilter
from .list_field_with_other_list_filter import ListFieldWithOtherListFilter
from .past_date_list_filter import PastDateListFilter
from .report_date_list_filter import ReportDateListFilter



# Generated by Django 5.2.10 on 2026-01-24 17:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0113_alter_historicalstock_container_qty_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalorderitem",
            name="container_qty",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="Number of units in the container(s). Must be the same for all containers in the record.",
                max_digits=20,
                null=True,
                verbose_name="Container unit quantity",
            ),
        ),
        migrations.AddField(
            model_name="orderitem",
            name="container_qty",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="Number of units in the container(s). Must be the same for all containers in the record.",
                max_digits=20,
                null=True,
                verbose_name="Container unit quantity",
            ),
        ),
        migrations.AlterField(
            model_name="historicalreceiveitem",
            name="container_qty",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="Number of units in the container(s). Must be the same for all containers in the record.",
                max_digits=20,
                null=True,
                verbose_name="Container unit quantity",
            ),
        ),
        migrations.AlterField(
            model_name="historicalreceiveitem",
            name="qty",
            field=models.DecimalField(
                decimal_places=2,
                max_digits=20,
                null=True,
                verbose_name="Number of containers",
            ),
        ),
        migrations.AlterField(
            model_name="historicalreceiveitem",
            name="unit_qty",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="number of containers x container unit qty, e.g. 10 x Bottle of 128 = 1280",
                max_digits=20,
                null=True,
                verbose_name="Total unit quantity",
            ),
        ),
        migrations.AlterField(
            model_name="receiveitem",
            name="container_qty",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="Number of units in the container(s). Must be the same for all containers in the record.",
                max_digits=20,
                null=True,
                verbose_name="Container unit quantity",
            ),
        ),
        migrations.AlterField(
            model_name="receiveitem",
            name="qty",
            field=models.DecimalField(
                decimal_places=2,
                max_digits=20,
                null=True,
                verbose_name="Number of containers",
            ),
        ),
        migrations.AlterField(
            model_name="receiveitem",
            name="unit_qty",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="number of containers x container unit qty, e.g. 10 x Bottle of 128 = 1280",
                max_digits=20,
                null=True,
                verbose_name="Total unit quantity",
            ),
        ),
    ]



from __future__ import annotations

from uuid import uuid4

from django.contrib import admin
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import gettext as _


@admin.action(description=_("Print stock report"))
def print_stock_report_action(modeladmin, request, queryset):
    if queryset.count() >= 1:
        session_uuid = str(uuid4())
        request.session[session_uuid] = list(queryset.values_list("pk", flat=True))
        url = reverse("edc_pharmacy:stock_report", kwargs={"session_uuid": session_uuid})
        return HttpResponseRedirect(url)
    return None



# Generated by Django 5.2.10 on 2026-01-24 20:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0116_remove_historicalorderitem_qty_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalorderitem",
            name="container_unit_qty",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Number of units in the container(s). Must be the same for all containers in the record.",
                max_digits=20,
                null=True,
                verbose_name="Container unit quantity",
            ),
        ),
        migrations.AlterField(
            model_name="historicalorderitem",
            name="item_qty_ordered",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Number of containers",
                max_digits=20,
                null=True,
                verbose_name="Number of containers ordered",
            ),
        ),
        migrations.AlterField(
            model_name="historicalorderitem",
            name="unit_qty_ordered",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Updated automatically (containers * unit_quantity per container)",
                max_digits=20,
                null=True,
                verbose_name="Unit quantity ordered",
            ),
        ),
        migrations.AlterField(
            model_name="historicalorderitem",
            name="unit_qty_pending",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Unit quantity ordered - Unit quantity received. Updated automatically",
                max_digits=20,
                null=True,
                verbose_name="Unit quantity pending",
            ),
        ),
        migrations.AlterField(
            model_name="historicalorderitem",
            name="unit_qty_received",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Updated automatically when units are received",
                max_digits=20,
                null=True,
                verbose_name="Unit quantity received",
            ),
        ),
        migrations.AlterField(
            model_name="historicalreceiveitem",
            name="item_qty_received",
            field=models.DecimalField(
                decimal_places=2,
                max_digits=20,
                null=True,
                verbose_name="Number of containers received",
            ),
        ),
        migrations.AlterField(
            model_name="historicalreceiveitem",
            name="unit_qty_received",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="number of containers x container unit qty, e.g. 10 x Bottle of 128 = 1280",
                max_digits=20,
                null=True,
                verbose_name="Unit quantity received",
            ),
        ),
        migrations.AlterField(
            model_name="orderitem",
            name="container_unit_qty",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Number of units in the container(s). Must be the same for all containers in the record.",
                max_digits=20,
                null=True,
                verbose_name="Container unit quantity",
            ),
        ),
        migrations.AlterField(
            model_name="orderitem",
            name="item_qty_ordered",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Number of containers",
                max_digits=20,
                null=True,
                verbose_name="Number of containers ordered",
            ),
        ),
        migrations.AlterField(
            model_name="orderitem",
            name="unit_qty_ordered",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Updated automatically (containers * unit_quantity per container)",
                max_digits=20,
                null=True,
                verbose_name="Unit quantity ordered",
            ),
        ),
        migrations.AlterField(
            model_name="orderitem",
            name="unit_qty_pending",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Unit quantity ordered - Unit quantity received. Updated automatically",
                max_digits=20,
                null=True,
                verbose_name="Unit quantity pending",
            ),
        ),
        migrations.AlterField(
            model_name="orderitem",
            name="unit_qty_received",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Updated automatically when units are received",
                max_digits=20,
                null=True,
                verbose_name="Unit quantity received",
            ),
        ),
        migrations.AlterField(
            model_name="receiveitem",
            name="item_qty_received",
            field=models.DecimalField(
                decimal_places=2,
                max_digits=20,
                null=True,
                verbose_name="Number of containers received",
            ),
        ),
        migrations.AlterField(
            model_name="receiveitem",
            name="unit_qty_received",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="number of containers x container unit qty, e.g. 10 x Bottle of 128 = 1280",
                max_digits=20,
                null=True,
                verbose_name="Unit quantity received",
            ),
        ),
    ]



from django.urls import path
from django.views.generic import RedirectView

from .admin_site import edc_sites_admin

app_name = "edc_sites"

urlpatterns = [
    path("admin/", edc_sites_admin.urls),
    path("", RedirectView.as_view(url=f"/{app_name}/admin/"), name="home_url"),
]



from __future__ import annotations

from typing import TYPE_CHECKING

from django.core.exceptions import ObjectDoesNotExist

from .exceptions import SingletonActionItemError
from .get_action_type import get_action_type

if TYPE_CHECKING:
    from .action import Action
    from .models import ActionItem


def create_action_item(
    action_cls: type[Action],
    subject_identifier: str = None,
    action_identifier: str | None = None,
    related_action_item: ActionItem | None = None,
    parent_action_item: ActionItem | None = None,
    priority: bool | None = None,
    using: str | None = None,
    site_id: int | None = None,
    **kwargs,
) -> ActionItem:
    action_item = None
    if action_cls.singleton:
        try:
            action_item = (
                action_cls.action_item_model_cls()
                .objects.using(using)
                .get(
                    action_type=get_action_type(action_cls),
                    subject_identifier=subject_identifier,
                )
            )
        except ObjectDoesNotExist:
            pass
        else:
            raise SingletonActionItemError(
                f"Action {action_cls.name} can only be "
                f"created once per subject. Got {subject_identifier}."
            )
    if not action_item:
        opts = dict(
            subject_identifier=subject_identifier,
            action_type=get_action_type(action_cls),
            action_identifier=action_identifier,
            linked_to_reference=False,
            priority=priority,
        )
        if parent_action_item:
            opts.update(parent_action_item=parent_action_item)
        if related_action_item:
            opts.update(related_action_item=related_action_item)
        if site_id:
            opts.update(site_id=site_id)
        action_item = action_cls.action_item_model_cls()(**opts)
        action_item.save(using=using)
        action_item.refresh_from_db()
    return action_item



# Generated by Django 6.0.1 on 2026-01-21 22:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_appointment", "0051_alter_appointment_revision_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="appointment",
            name="appt_timing",
            field=models.CharField(
                choices=[
                    ("ontime", "On time (within window period)"),
                    ("missed", "Missed"),
                    (
                        "extended",
                        "Extended (extended window period for final appointment)",
                    ),
                    ("N/A", "Not applicable"),
                ],
                default="ontime",
                help_text="If late, you may also be required to complete a protocol incident report. Extended window period may be allowed for the final appointment. Refer to the protocol documentation for the allowed window periods of scheduled appointments.",
                max_length=25,
                verbose_name="Timing",
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="appt_timing",
            field=models.CharField(
                choices=[
                    ("ontime", "On time (within window period)"),
                    ("missed", "Missed"),
                    (
                        "extended",
                        "Extended (extended window period for final appointment)",
                    ),
                    ("N/A", "Not applicable"),
                ],
                default="ontime",
                help_text="If late, you may also be required to complete a protocol incident report. Extended window period may be allowed for the final appointment. Refer to the protocol documentation for the allowed window periods of scheduled appointments.",
                max_length=25,
                verbose_name="Timing",
            ),
        ),
    ]



from edc_dx import get_diagnosis_labels


def get_clinical_review_baseline_cond_fieldset(
    cond: str, title: str = None
) -> tuple[str, dict]:
    if not title:
        title = cond.upper()
    return (
        title,
        {"fields": (f"{cond}_dx", f"{cond}_dx_at_screening")},
    )


def get_clinical_review_baseline_cond_fieldsets() -> tuple[tuple]:
    fieldsets = ()
    for prefix, label in get_diagnosis_labels().items():
        fieldsets = fieldsets + (
            get_clinical_review_baseline_cond_fieldset(cond=prefix.lower(), title=label),
        )
    return fieldsets


def get_clinical_review_cond_fieldset(cond: str, title: str = None) -> tuple[str, dict]:
    if not title:
        title = cond.upper()
    return (
        title,
        {
            "fields": (
                f"{cond}_test",
                f"{cond}_test_date",
                f"{cond}_reason",
                f"{cond}_reason_other",
                f"{cond}_dx",
            )
        },
    )


def get_clinical_review_cond_fieldsets() -> tuple[tuple]:
    fieldsets = ()
    for prefix, label in get_diagnosis_labels().items():
        fieldsets = fieldsets + (
            get_clinical_review_cond_fieldset(cond=prefix.lower(), title=label),
        )
    return fieldsets



import urllib

from django.http.response import HttpResponseRedirect
from django.urls.base import reverse
from django.utils.text import slugify
from django.views.generic.base import TemplateView

from edc_dashboard.url_names import url_names

from ...dashboard_templates import dashboard_templates


class InvalidPostError(Exception):
    pass


class ActionViewError(Exception):
    pass


class ActionView(TemplateView):
    """A view for lab "actions" such as receive, process,
    aliquot, etc.
    """

    form_action_selected_items_name: str = "selected_items"
    post_action_url: str = None  # key exists in url_names
    redirect_querystring: dict = {}  # noqa: RUF012
    template_name: str = dashboard_templates.get("home_template")
    valid_form_actions: tuple[str, ...] = ()

    navbar_name: str = "specimens"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._selected_items = []
        self.action = None

    def process_form_action(self, request=None):
        """Override to conditionally handle the action POST attr."""
        pass

    @property
    def selected_items(self):
        """Returns a list of selected listboard items."""
        if not self._selected_items:
            self._selected_items = (
                self.request.POST.getlist(self.form_action_selected_items_name) or []
            )
            self._selected_items = [x for x in self._selected_items if x]
        return self._selected_items

    @property
    def url_kwargs(self):
        """Returns the default dictionary to reverse the post url."""
        return {}

    def post(self, request, *args, **kwargs):
        """Process the form "action" then redirect."""
        action = slugify(self.request.POST.get("action", "").lower())
        if action not in self.valid_form_actions:
            raise InvalidPostError(f"Invalid form action in POST. Got {action}")
        self.action = action
        self.process_form_action(request=request)
        url_name = url_names.get(self.post_action_url)
        url = reverse(url_name, kwargs=self.url_kwargs)
        if self.redirect_querystring:
            url = f"{url}?{urllib.parse.urlencode(self.redirect_querystring)}"
        return HttpResponseRedirect(url)



# Generated by Django 5.2.10 on 2026-01-25 23:12

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0131_storagebinitemproxy_storagebinproxy"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalstoragebin",
            name="container",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_pharmacy.container",
            ),
        ),
        migrations.AddField(
            model_name="storagebin",
            name="container",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_pharmacy.container",
            ),
        ),
    ]



from django.db import models

from edc_action_item.managers import (
    ActionIdentifierModelManager,
    ActionIdentifierSiteManager,
)
from edc_action_item.models import ActionModelMixin
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_sites.model_mixins import SiteModelMixin

from ...constants import AE_FOLLOWUP_ACTION
from .ae_followup_fields_model_mixin import AeFollowupFieldsModelMixin
from .ae_followup_methods_model_mixin import AeFollowupMethodsModelMixin


class AeFollowupModelMixin(
    SiteModelMixin,
    ActionModelMixin,
    NonUniqueSubjectIdentifierFieldMixin,
    AeFollowupFieldsModelMixin,
    AeFollowupMethodsModelMixin,
    models.Model,
):
    action_name = AE_FOLLOWUP_ACTION

    objects = ActionIdentifierModelManager()

    on_site = ActionIdentifierSiteManager()

    class Meta(
        SiteModelMixin.Meta,
        NonUniqueSubjectIdentifierFieldMixin.Meta,
        ActionModelMixin.Meta,
    ):
        abstract = True
        verbose_name = "AE Follow-up Report"
        indexes = (
            *NonUniqueSubjectIdentifierFieldMixin.Meta.indexes,
            *ActionModelMixin.Meta.indexes,
            models.Index(fields=["subject_identifier", "action_identifier", "site", "id"]),
        )



from django.conf import settings
from django.contrib import admin
from django.contrib.admin import sites
from django.contrib.auth import get_permission_codename
from django.urls import NoReverseMatch, reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django_audit_fields.admin import ModelAdminAuditFieldsMixin, audit_fieldset_tuple
from django_revision.modeladmin_mixin import ModelAdminRevisionMixin

from edc_list_data.model_mixins import ListModelMixin
from edc_model_admin.history import SimpleHistoryAdmin
from edc_model_admin.mixins import (
    ModelAdminFormInstructionsMixin,
    ModelAdminInstitutionMixin,
    ModelAdminNextUrlRedirectMixin,
    ModelAdminRedirectOnDeleteMixin,
    ModelAdminReplaceLabelTextMixin,
    TemplatesModelAdminMixin,
)
from edc_notification.modeladmin_mixins import NotificationModelAdminMixin

from ..admin_site import edc_data_manager_admin
from ..forms import DataDictionaryForm
from ..models import DataDictionary
from ..populate_data_dictionary import populate_data_dictionary


@admin.register(DataDictionary, site=edc_data_manager_admin)
class DataDictionaryAdmin(
    TemplatesModelAdminMixin,
    ModelAdminNextUrlRedirectMixin,  # add
    NotificationModelAdminMixin,
    ModelAdminFormInstructionsMixin,  # add
    ModelAdminRevisionMixin,  # add
    ModelAdminInstitutionMixin,  # add
    ModelAdminRedirectOnDeleteMixin,
    ModelAdminReplaceLabelTextMixin,
    ModelAdminAuditFieldsMixin,
    SimpleHistoryAdmin,
):

    form = DataDictionaryForm

    fieldsets = (
        [
            None,
            {
                "fields": (
                    "model",
                    "model_verbose_name",
                    "number",
                    "prompt",
                    "field_name",
                    "field_type",
                    "help_text",
                    "default_value",
                    "nullable",
                    "max_length",
                    "max_digits",
                    "decimal_places",
                    "active",
                )
            },
        ],
        audit_fieldset_tuple,
    )

    ordering = ("active", "model", "number")

    list_filter = ("active", "model", "field_type", "nullable")

    list_display = (
        "form_title",
        "field_name",
        "changelist_link",
        "active",
        "number",
        "question_text",
        "field_type",
        "default_value",
        "help_text",
        "nullable",
        "max_length",
        "max_digits",
        "decimal_places",
    )

    actions = ("populate_data_dictionary_action",)

    search_fields = (
        "number",
        "prompt",
        "field_name",
        "model",
        "help_text",
        "model_verbose_name",
    )

    readonly_fields = [
        "model",
        "model_verbose_name",
        "number",
        "prompt",
        "field_name",
        "field_type",
        "help_text",
        "default_value",
        "nullable",
        "max_length",
        "max_digits",
        "decimal_places",
        "active",
    ]

    @admin.display(description="Changelist")
    def changelist_link(self, obj):
        for site in sites.all_sites.data:
            for modeladmin in site()._registry.values():
                if modeladmin.model._meta.label_lower == obj.model:
                    model_string = obj.model.replace(".", "_")
                    try:
                        url = reverse(f"{site().name}:{model_string}_changelist")
                    except NoReverseMatch:
                        pass
                    else:
                        return format_html(
                            '<a href="{url}">Changelist</a>',
                            url=mark_safe(url),  # nosec B703 B308
                        )
        return None

    @staticmethod
    def form_title(obj):
        return obj.model_verbose_name

    @staticmethod
    def question_text(obj):
        return format_html(
            "{}",
            mark_safe(obj.prompt),  # nosec B703, B308
        )

    def populate_data_dictionary_action(self, request, queryset):
        DataDictionary.objects.update(active=False)
        for site in sites.all_sites.data:
            for model_admin in site()._registry.values():
                form = model_admin.get_form(request)
                model = model_admin.model
                if (
                    model._meta.app_label in settings.DATA_DICTIONARY_APP_LABELS
                    and not issubclass(model, (ListModelMixin,))
                ):
                    populate_data_dictionary(form=form, model=model)

    populate_data_dictionary_action.allowed_permissions = ("refresh",)
    populate_data_dictionary_action.short_description = "Refresh data dictionary"

    def has_refresh_permission(self, request):
        opts = self.opts
        codename = get_permission_codename("delete", opts)
        return request.user.has_perm("%s.%s" % (opts.app_label, codename))



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from django.core.management.color import color_style
from django.urls import NoReverseMatch
from django.urls.base import reverse

from edc_dashboard.url_names import InvalidDashboardUrlName, url_names

if TYPE_CHECKING:
    from django.contrib.auth.models import User

style = color_style()


@dataclass
class NavbarItem:
    """A class that represents a single item on a navbar."""

    name: str = field(default=None, compare=True)
    title: str = field(default=None)
    label: str | None = field(default=None)
    codename: str = field(default=None)
    url_names_key: str | None = field(
        default=None
    )  # must be valid key in url_names dictionary
    url_with_namespace: str | None = field(default=None)
    url_without_namespace: str | None = field(default=None)
    fa_icon: str | None = field(default=None)
    disabled: str | None = field(default="disabled")

    active: bool = field(default=None)

    template_name: str = field(
        default="edc_navbar/navbar_item.html",
        repr=False,
    )

    def __post_init__(self):
        self.title = self.title or self.label or self.name.title()  # the anchor title
        if self.url_with_namespace and ":" not in self.url_with_namespace:
            raise InvalidDashboardUrlName(
                f"Invalid url_with_namespace. Got {self.url_with_namespace}"
            )

    def get_url(self, raise_exception: bool | None = None) -> str | None:
        url = (
            self.url_without_namespace
            or self.url_with_namespace
            or url_names.get(self.url_names_key)
        )
        try:
            url = reverse(url)
        except NoReverseMatch as e:
            if raise_exception:
                errmsg = (
                    f"Reverse for Navbar url not found. Tried {url}. "
                    f"See NavbarItem(name={self.name}"
                )
                if self.url_without_namespace == url:
                    errmsg = f"{errmsg}, url_without_namespace={self.url_without_namespace})."
                elif self.url_with_namespace == url:
                    errmsg = f"{errmsg}, url_with_namespace={self.url_with_namespace})."
                elif url_names.get(self.url_names_key) == url:
                    errmsg = f"{errmsg}, url_names_key={self.url_names_key})."
                raise NoReverseMatch(errmsg) from e
            url = None
        return url

    def set_disabled(self, user: User | None = None):
        if user and user.has_perm(self.codename):
            self.disabled = ""



from dateutil.relativedelta import relativedelta
from django.conf import settings
from django.utils.translation import gettext as _

from edc_utils import (
    ceil_secs,
    convert_php_dateformat,
    floor_secs,
    formatted_date,
    to_local,
)

from ..exceptions import (
    ScheduledVisitWindowError,
    UnScheduledVisitWindowError,
)
from ..utils import get_enforce_window_period_enabled


class Window:
    def __init__(
        self,
        name=None,
        dt=None,
        baseline_timepoint_datetime=None,
        timepoint_datetime=None,
        visit=None,
        next_visit=None,
        visit_code_sequence=None,
    ):
        self.name = name

        # convert dates to local tzinfo
        self.dt = to_local(dt)
        self.timepoint_datetime = to_local(timepoint_datetime)
        self.baseline_timepoint_datetime = to_local(baseline_timepoint_datetime)

        self.visit = visit
        self.visit.timepoint_datetime = self.timepoint_datetime

        self.next_visit = next_visit
        if self.next_visit:
            self.next_visit.timepoint_datetime = (
                self.baseline_timepoint_datetime + self.next_visit.rbase
            )

        self.visit_code_sequence = visit_code_sequence

    @property
    def datetime_in_window(self):
        if get_enforce_window_period_enabled():
            if not self.dt:
                raise UnScheduledVisitWindowError("Invalid datetime")
            if self.is_scheduled_visit:
                self.raise_for_scheduled_not_in_window()
            else:
                self.raise_for_unscheduled_not_in_window()
        return True

    @property
    def is_scheduled_visit(self):
        return self.visit_code_sequence == 0 or self.visit_code_sequence is None

    def get_window_gap_days(self) -> int:
        days = 0
        if self.visit.add_window_gap_to_lower and self.next_visit:
            days = abs(
                (self.timepoint_datetime + self.visit.rupper)
                - (self.timepoint_datetime - self.next_visit.rlower)
            ).days
        return days

    def raise_for_scheduled_not_in_window(self):
        """Returns the datetime if it falls within the
        window period for a scheduled `visit` otherwise
        raises an exception.

        In this case, `visit` is the object from schedule and
        not a model instance.
        """

        gap_days = self.get_window_gap_days()
        lower = floor_secs(to_local(self.visit.dates.lower) - relativedelta(days=gap_days))
        upper = ceil_secs(to_local(self.visit.dates.upper))
        if not (lower <= floor_secs(to_local(self.dt)) <= upper):
            lower_date = to_local(lower).strftime(
                convert_php_dateformat(settings.SHORT_DATETIME_FORMAT)
            )
            upper_date = to_local(upper).strftime(
                convert_php_dateformat(settings.SHORT_DATETIME_FORMAT)
            )
            dt = to_local(self.dt).strftime(
                convert_php_dateformat(settings.SHORT_DATETIME_FORMAT)
            )

            raise ScheduledVisitWindowError(
                "Invalid. Date falls outside of the "
                f"window period for this `scheduled` visit. "
                f"Expected a date between {lower_date} "
                f"and {upper_date} for {self.visit.code}. Got `{dt}`."
            )

    def raise_for_unscheduled_not_in_window(self):
        """Returns the datetime if it falls within the
        window period for an unscheduled `visit` otherwise
        raises an exception.

        Window period for an unscheduled date is anytime
        on or after the scheduled date and before the projected
        lower bound of the next visit.

        In this case, `visit` is the object from schedule and
        not a model instance.
        """
        formatted_dt = formatted_date(self.dt)
        if self.next_visit:
            in_window = floor_secs(to_local(self.dt)) < floor_secs(
                to_local(self.next_visit.timepoint_datetime - self.next_visit.rlower)
            )
            msg = _(
                "Invalid datetime. Falls outside of the "
                "window period for this `unscheduled` visit. "
                "Expected a datetime before the next visit. "
                "Next visit is `%(next_visit_code)s` expected any time "
                "from `%(dt_lower)s`."
                "Got `%(visit_code)s`@`%(dt)s`. "
            ) % dict(
                next_visit_code=self.next_visit.code,
                dt_lower=formatted_date(
                    self.next_visit.timepoint_datetime - self.next_visit.rlower
                ),
                visit_code=self.visit.code,
                dt=formatted_dt,
            )
        else:
            in_window = floor_secs(to_local(self.dt)) < floor_secs(
                to_local(
                    self.visit.timepoint_datetime
                    + (self.visit.rupper_extended or self.visit.rupper)
                )
            )
            msg = _(
                "Invalid datetime. Falls outside of the "
                "window period for this `unscheduled` visit. "
                "Expected a datetime before `%(dt_upper)s`."
                "Got `%(visit_code)s`@`%(dt)s`. "
            ) % dict(
                dt_upper=formatted_date(
                    to_local(
                        self.visit.timepoint_datetime
                        + (self.visit.rupper_extended or self.visit.rupper)
                    )
                ),
                visit_code=self.visit.code,
                dt=formatted_dt,
            )
        if not in_window:
            raise UnScheduledVisitWindowError(msg)



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from django.core.exceptions import ValidationError
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _
from edc_appointment.form_validator_mixins import WindowPeriodFormValidatorMixin
from edc_appointment.utils import allow_extended_window_period
from edc_form_validators import INVALID_ERROR, FormValidator
from edc_registration import get_registered_subject_model_cls
from edc_sites.form_validator_mixin import SiteFormValidatorMixin
from edc_utils import floor_secs, formatted_datetime
from edc_utils.date import to_local
from edc_visit_schedule.exceptions import ScheduledVisitWindowError
from edc_visit_tracking.modelform_mixins import get_related_visit

from .crf_form_validator_mixins import CrfFormValidatorMixin

if TYPE_CHECKING:
    from edc_appointment.models import Appointment
    from edc_visit_tracking.model_mixins import VisitModelMixin


class CrfFormValidatorError(Exception):
    pass


class CrfFormValidator(
    WindowPeriodFormValidatorMixin,
    CrfFormValidatorMixin,
    SiteFormValidatorMixin,
    FormValidator,
):
    """Form validator for CRfs.

    CRFs have a FK to related_visit_model_cls.
    """

    report_datetime_field_attr = "report_datetime"

    def _clean(self) -> None:
        self.validate_crf_report_datetime()
        super()._clean()

    @property
    def registered_subject(self):
        return get_registered_subject_model_cls().objects.get(
            subject_identifier=self.subject_identifier
        )

    @property
    def screening_identifier(self) -> str:
        return self.registered_subject.screening_identifier

    @property
    def gender(self):
        return self.registered_subject.gender

    @property
    def dob(self):
        return self.registered_subject.dob

    def validate_crf_report_datetime(self) -> None:
        if self.report_datetime:
            # falls within appointment's window period
            self.validate_crf_datetime_in_window_period()
            # not before consent date
            consent_datetime = self.get_consent_datetime_or_raise()
            if floor_secs(self.report_datetime) < floor_secs(consent_datetime):
                msg = _("Invalid. Cannot be before date of consent. Participant consented on")
                formatted_date = formatted_datetime(to_local(consent_datetime))
                err_message = format_lazy(
                    "{msg} {formatted_date}", msg=msg, formatted_date=formatted_date
                )
                self.raise_validation_error(
                    {self.report_datetime_field_attr: err_message},
                    INVALID_ERROR,
                )

    def validate_crf_datetime_in_window_period(self) -> None:
        try:
            self.datetime_in_window_or_raise(
                self.appointment,
                self.report_datetime,
                self.report_datetime_field_attr,
            )
        except (ScheduledVisitWindowError, ValidationError):
            if not allow_extended_window_period(
                self.appointment.appt_timing,
                self.report_datetime,
                self.appointment,
            ):
                raise

    @property
    def appointment(self) -> Appointment:
        try:
            appointment = self.related_visit.appointment
        except AttributeError as e:
            msg = _("is required")
            verbose_name = self.related_visit._meta.verbose_name
            self.raise_validation_error(
                format_lazy("{verbose_name} {msg}.", msg=msg, verbose_name=verbose_name),
                INVALID_ERROR,
                exc=e,
            )
        return appointment

    @property
    def related_visit_model_attr(self) -> str:
        """Returns the field name attr if the  related_visit.

        Note: during testing, `self.instance` and `self.model` are
        set on the FormValidator class by a modelform mixin. If you
        are testing the FormValidator in isolation, add these
        values manually on behalf of the modelform. See also
        `related_visit`.
        """
        try:
            return self.instance.related_visit_model_attr()
        except AttributeError:
            return self.model.related_visit_model_attr()

    @property
    def related_visit(self) -> VisitModelMixin | None:
        """Returns a related_visit (subject visit) model instance
        or None.
        """
        try:
            return get_related_visit(
                self, related_visit_model_attr=self.related_visit_model_attr
            )
        except AttributeError as e:
            raise CrfFormValidatorError(f"{e}. See {self.__class__}") from e

    def validate_datetime_against_report_datetime(self, field: str) -> None:
        """Datetime cannot be after report_datetime"""
        if (
            self.cleaned_data.get(field)
            and floor_secs(self.report_datetime)
            and floor_secs(self.cleaned_data.get(field)) > floor_secs(self.report_datetime)
        ):
            msg = _("Cannot be after report datetime")
            self.raise_validation_error({field: msg}, INVALID_ERROR)

    def validate_date_against_report_datetime(self, field: str) -> None:
        """Date cannot be after report_datetime"""
        if (
            self.cleaned_data.get(field)
            and self.report_datetime
            and self.cleaned_data.get(field) > self.report_datetime.date()
        ):
            self.raise_validation_error(
                {field: _("Cannot be after report datetime")}, INVALID_ERROR
            )

    @property
    def subject_visit(self) -> VisitModelMixin:
        warnings.warn(
            "The subject_visit attribute is deprecated in favor of related_visit.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.related_visit



from .ae_followup_action import AeFollowupAction
from .ae_initial_action import AeInitialAction
from .ae_susar_action import AeSusarAction
from .ae_tmg_action import AeTmgAction
from .death_report_action import DeathReportAction
from .death_report_tmg_action import DeathReportTmgAction
from .death_report_tmg_second_action import DeathReportTmgSecondAction
from .hospitalization_action import HospitalizationAction



from django.apps import apps as django_apps
from django.core.management.base import BaseCommand, CommandError

from edc_form_runners.exceptions import FormRunnerError
from edc_form_runners.run_form_runners import run_form_runners


class Command(BaseCommand):
    help = "Run form runners"

    def add_arguments(self, parser):
        parser.add_argument(
            "-a",
            "--app",
            dest="app_labels",
            default="",
            help="if more than one separate by comma",
        )

        parser.add_argument(
            "-m",
            "--model",
            dest="model_names",
            default="",
            help="model name in label_lower format, if more than one separate by comma",
        )

        parser.add_argument(
            "-s",
            "--skip_model",
            dest="skip_model_names",
            default="",
            help="model to skip in label_lower format, if more than one separate by comma",
        )

        parser.add_argument(
            "--debug",
            dest="debug",
            action="store_true",
            default=False,
            help="debug mode",
        )

    def handle(self, *args, **options):
        debug = options["debug"]

        app_labels = options["app_labels"] or []
        if app_labels:
            app_labels = options["app_labels"].split(",")

        model_names = options["model_names"] or []
        if model_names:
            model_names = options["model_names"].split(",")

        skip_model_names = options["skip_model_names"] or []
        if skip_model_names:
            skip_model_names = options["skip_model_names"].split(",")

        if app_labels and model_names:
            raise CommandError(
                "Either provide the `app label` or `model name(s)` but not both. "
                f"Got {app_labels} and {model_names}."
            )

        # if app_labels:
        #     for app_config in django_apps.get_app_configs():
        #         if app_config.name in app_labels:
        #             for model_cls in app_config.get_models():
        #                 if not model_cls._meta.label_lower.split(".")[1].startswith(
        #                     "historical"
        #                 ):
        #                     model_names.append(model_cls._meta.label_lower)
        #
        # model_names = [m for m in model_names if m not in skip_model_names]
        model_names = get_model_names(app_labels, model_names, skip_model_names)

        try:
            run_form_runners(model_names=model_names)
        except FormRunnerError as e:
            if debug:
                raise
            raise CommandError(e)


def get_model_names(app_labels, model_names, skip_model_names):
    if app_labels:
        for app_config in django_apps.get_app_configs():
            if app_config.name in app_labels:
                for model_cls in app_config.get_models():
                    if not model_cls._meta.label_lower.split(".")[1].startswith("historical"):
                        model_names.append(model_cls._meta.label_lower)

    return [m for m in model_names if m not in skip_model_names]



# Generated by Django 6.0 on 2025-03-04 21:03

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0058_stockrequestproxy_alter_stockproxy_options"),
    ]

    operations = [
        migrations.DeleteModel(
            name="StockRequestProxy",
        ),
    ]



from .list_filters import NoteStatusListFilter
from .note_modeladmin_mixin import NoteModelAdminMixin
from .on_study_missing_values_modeladmin_mixin import (
    OnStudyMissingValuesModelAdminMixin,
)
from .qa_report_modeladmin_mixin import QaReportModelAdminMixin



from uuid import uuid4

from celery.states import PENDING
from django.apps import apps as django_apps
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.safestring import mark_safe
from django.views.generic import TemplateView
from django_pandas.io import read_frame

from edc_dashboard.view_mixins import EdcViewMixin
from edc_navbar import NavbarViewMixin
from edc_protocol.view_mixins import EdcProtocolViewMixin
from edc_utils.celery import celery_is_active, get_task_result

from ..analytics import get_next_scheduled_visit_for_subjects_df
from ..models import StockRequest, StockRequestItem
from ..utils import bulk_create_stock_request_items, get_instock_and_nostock_data


@method_decorator(login_required, name="dispatch")
class PrepareAndReviewStockRequestView(
    EdcViewMixin, NavbarViewMixin, EdcProtocolViewMixin, TemplateView
):
    template_name: str = "edc_pharmacy/stock/stock_request.html"
    navbar_name = settings.APP_NAME
    navbar_selected_item = "pharmacy"

    def get_context_data(self, **kwargs):
        stock_request = StockRequest.objects.get(pk=self.kwargs.get("stock_request"))
        df = get_next_scheduled_visit_for_subjects_df(stock_request)

        # get unallocated stock that appears in a stock request for this location
        df_unallocated_request_items = (
            read_frame(
                StockRequestItem.objects.values(
                    "stock_request__request_identifier",
                    "registered_subject__subject_identifier",
                ).filter(
                    stock_request__location=stock_request.location, allocation__isnull=True
                )
            )
            .rename(
                columns={
                    "registered_subject__subject_identifier": "subject_identifier",
                    "stock_request__request_identifier": "request_identifier",
                },
            )
            .reset_index(drop=True)
        )
        df_unallocated_request_items = df_unallocated_request_items.merge(
            df[["subject_identifier", "next_visit_code", "next_appt_datetime"]],
            on="subject_identifier",
            how="left",
        )
        df_unallocated_request_items = df_unallocated_request_items[
            df_unallocated_request_items.next_visit_code.notna()
        ]
        df_unallocated_request_items = df_unallocated_request_items.sort_values(
            by=["subject_identifier"]
        ).reset_index()

        # exclude unallocated subjects from appts
        df = df[
            ~df.subject_identifier.isin(df_unallocated_request_items.subject_identifier)
        ].reset_index(drop=True)

        kwargs.update(
            stock_request=stock_request,
            stock_request_items_exist=stock_request.stockrequestitem_set.all().exists(),
            source_model_name=self.model_cls._meta.verbose_name_plural,
            source_changelist_url=self.source_changelist_url,
            rows=0,
            subjects=[],
        )

        if df.empty:
            messages.add_message(
                self.request,
                messages.ERROR,
                (
                    f"No future subject appointments found for {stock_request.location} "
                    f"with this cutoff date. (Site {stock_request.location.site.id})."
                ),
            )
        elif getattr(get_task_result(stock_request), "status", "") == PENDING:
            messages.add_message(
                self.request,
                messages.ERROR,
                (
                    f"Stock request {stock_request.request_identifier} is still processing. "
                    "Please click cancel and check the status column."
                ),
            )
        else:
            df_instock, df_nostock = get_instock_and_nostock_data(stock_request, df)
            session_uuid = str(uuid4())
            nostock_dict = df_nostock[
                [
                    "subject_identifier",
                    "registered_subject_id",
                    "next_visit_code",
                    "next_appt_datetime",
                ]
            ].to_dict("list")
            nostock_dict["stock_qty"] = 0.0
            self.request.session[session_uuid] = nostock_dict

            stock_request_items_exist = stock_request.stockrequestitem_set.all().exists()
            if stock_request_items_exist:
                messages.add_message(
                    self.request,
                    messages.ERROR,
                    message=(
                        f"Stock request items already exist for "
                        f"{stock_request._meta.verbose_name} "
                        f"{stock_request.request_identifier}. "
                        "Create has been disabled."
                    ),
                )
            kwargs.update(
                rows=len(df_nostock),
                subjects=df_nostock.subject_identifier.nunique(),
                subjects_excluded_by_stock=len(df_instock.subject_identifier.unique()),
                subjects_excluded_by_request=len(
                    df_unallocated_request_items.subject_identifier.unique()
                ),
                nostock_table=mark_safe(  # noqa: S308
                    df_nostock.to_html(
                        columns=[
                            "subject_identifier",
                            "next_visit_code",
                            "next_appt_datetime",
                        ],
                        index=True,
                        border=0,
                        classes="table table-striped",
                        table_id="my_table",
                    )
                ),
                instock_table=mark_safe(  # noqa: S308
                    df_instock.to_html(
                        columns=[
                            "subject_identifier",
                            "next_visit_code",
                            "next_appt_datetime",
                            "code",
                        ],
                        index=True,
                        border=0,
                        classes="table table-striped",
                        table_id="in_stock_table",
                    )
                ),
                unallocated_table=mark_safe(  # noqa: S308
                    df_unallocated_request_items.to_html(
                        columns=[
                            "subject_identifier",
                            "next_visit_code",
                            "next_appt_datetime",
                            "request_identifier",
                        ],
                        index=True,
                        border=0,
                        classes="table table-striped",
                        table_id="unallocated_table",
                    )
                ),
                session_uuid=session_uuid,
            )
        return super().get_context_data(**kwargs)

    @property
    def source_changelist_url(self):
        return reverse("edc_pharmacy_admin:edc_pharmacy_stockrequest_changelist")

    @property
    def model_cls(self):
        return django_apps.get_model("edc_pharmacy.stocktransfer")

    def post(self, request, *args, **kwargs):  # noqa: ARG002
        session_uuid = request.POST.get("session_uuid")
        stock_request = StockRequest.objects.get(pk=request.POST.get("stock_request"))
        if not request.POST.get("cancel") and session_uuid:
            nostock_dict = request.session[session_uuid]
            if session_uuid:
                del request.session[session_uuid]

            task_id = None
            if not celery_is_active():
                bulk_create_stock_request_items(
                    stock_request.pk, nostock_dict, user_created=request.user.username
                )
            else:
                task = bulk_create_stock_request_items.delay(
                    stock_request.pk, nostock_dict, user_created=request.user.username
                )
                task_id = getattr(task, "id", None)
            obj = StockRequest.objects.get(pk=request.POST.get("stock_request"))
            obj.task_id = task_id
            obj.save(update_fields=["task_id"])

            messages.add_message(
                request,
                messages.SUCCESS,
                (
                    f"Successfully created items for Stock Request "
                    f"{stock_request.request_identifier}"
                ),
            )
            url = f"{self.source_changelist_url}?q={stock_request.request_identifier}"
        else:
            if session_uuid:
                del request.session[session_uuid]
            messages.add_message(
                request,
                messages.INFO,
                "Cancelled. No stock request items were created.",
            )
            url = f"{self.source_changelist_url}"
        return HttpResponseRedirect(url)



# Generated by Django 5.2.10 on 2026-01-25 02:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0124_alter_historicalrepackrequest_repack_item_qty_and_more"),
    ]

    operations = [
        migrations.RenameField(
            model_name="historicalrepackrequest",
            old_name="repack_item_qty",
            new_name="item_qty_repack",
        ),
        migrations.RenameField(
            model_name="repackrequest",
            old_name="repack_item_qty",
            new_name="item_qty_repack",
        ),
        migrations.RenameField(
            model_name="historicalrepackrequest",
            old_name="processed_qty",
            new_name="item_qty_processed",

        ),
        migrations.RenameField(
            model_name="repackrequest",
            old_name="processed_qty",
            new_name="item_qty_processed",

        ),
    ]



from edc_protocol.research_protocol_config import ResearchProtocolConfig
from reportlab.graphics.barcode.widgets import BarcodeStandard39
from reportlab.graphics.shapes import Drawing, String

from ..models import Stock
from ..utils import format_qty


def draw_bulk_stock_label_code39(
    label: Drawing,
    width: int | float,
    height: int | float,
    obj: Stock,
) -> Drawing:
    """Callable to draw a single study medication label given a model
    instance `obj`
    """
    br = BarcodeStandard39(
        humanReadable=True, checksum=False, barHeight=30, barWidth=0.7, gap=1.7
    )
    br.value = obj.code
    br.x = width - 100
    br.y = height - 40
    label.add(br)
    label.add(String(15, height - 20, ResearchProtocolConfig().protocol_name, fontSize=10))
    qty = format_qty(obj.container_unit_qty, obj.container)

    label.add(String(15, height - 40, f"{qty} tabs", fontSize=10))
    label.add(String(15, height - 60, f"Lot: {obj.lot.lot_no}", fontSize=10))
    label.add(String(15, height - 80, f"Expires: {obj.lot.expiration_date}", fontSize=10))
    product = obj.product.formulation.get_description_with_assignment(obj.product.assignment)
    label.add(String(15, height - 100, f"{product}", fontSize=10))
    return label



from clinicedc_constants import CANCEL
from django import forms
from edc_registration.models import RegisteredSubject

from ...models import Allocation, StockRequest


class StockRequestForm(forms.ModelForm):
    def clean(self):
        cleaned_data = super().clean()

        if cleaned_data.get("subject_identifiers") and cleaned_data.get(
            "excluded_subject_identifiers"
        ):
            raise forms.ValidationError(
                "Cannot include and exclude subject identifiers in the same request."
            )

        self.clean_dates(cleaned_data)

        if cleaned_data.get("subject_identifiers") and cleaned_data.get("location"):
            subject_identifiers = cleaned_data.get("subject_identifiers").split("\n")
            subject_identifiers = [s.strip() for s in subject_identifiers]
            self.cleaned_data["subject_identifiers"] = "\n".join(subject_identifiers)
            if RegisteredSubject.objects.values("subject_identifier").filter(
                subject_identifier__in=subject_identifiers,
                site_id=cleaned_data.get("location").site_id,
            ).count() != len(subject_identifiers):
                raise forms.ValidationError(
                    {"subject_identifiers": "Not all subject identifiers are valid."}
                )
        if cleaned_data.get("excluded_subject_identifiers") and cleaned_data.get("location"):
            subject_identifiers = cleaned_data.get("excluded_subject_identifiers").split("\n")
            subject_identifiers = [s.strip() for s in subject_identifiers]
            self.cleaned_data["excluded_subject_identifiers"] = "\n".join(subject_identifiers)
            if RegisteredSubject.objects.values("subject_identifier").filter(
                subject_identifier__in=subject_identifiers,
                site_id=cleaned_data.get("location").site_id,
            ).count() != len(subject_identifiers):
                raise forms.ValidationError(
                    {
                        "excluded_subject_identifiers": (
                            "Not all subject identifiers are valid. Type one subject per line."
                        )
                    }
                )

        if (
            cleaned_data.get("container")
            and cleaned_data.get("containers_per_subject")
            and cleaned_data.get("containers_per_subject")
            > cleaned_data.get("container").max_items_per_subject
        ):
            raise forms.ValidationError(
                {
                    "containers_per_subject": (
                        f"May not exceed {cleaned_data.get('container').max_items_per_subject}. "
                        "See 'max per subject' for this container"
                    )
                }
            )

        if not self.instance.id and cleaned_data.get("cancel") == CANCEL:
            raise forms.ValidationError("Leave this blank")
        if (
            cleaned_data.get("cancel") == CANCEL
            and Allocation.objects.filter(
                stock_request_item__stock_request=self.instance
            ).exists()
        ):
            raise forms.ValidationError(
                "May not be cancelled. Stock has been allocated for this request"
            )
        return cleaned_data

    @staticmethod
    def clean_dates(cleaned_data):
        if (
            cleaned_data.get("request_datetime")
            and cleaned_data.get("cutoff_datetime")
            and cleaned_data.get("cutoff_datetime") < cleaned_data.get("request_datetime")
        ):
            raise forms.ValidationError(
                {"cutoff_datetime": "Invalid. Must after the request date"}
            )
        if (
            cleaned_data.get("request_datetime")
            and cleaned_data.get("cutoff_datetime")
            and cleaned_data.get("cutoff_datetime").date()
            == cleaned_data.get("request_datetime").date()
        ):
            raise forms.ValidationError(
                {"cutoff_datetime": "Invalid. Must be at least 1 day after the request date"}
            )
        if (
            cleaned_data.get("start_datetime")
            and cleaned_data.get("request_datetime")
            and cleaned_data.get("start_datetime").date()
            > cleaned_data.get("request_datetime").date()
        ):
            raise forms.ValidationError(
                {"start_datetime": "Invalid.  Must on or before the request date"}
            )
        if (
            cleaned_data.get("start_datetime")
            and cleaned_data.get("cutoff_datetime")
            and cleaned_data.get("start_datetime").date()
            >= cleaned_data.get("cutoff_datetime").date()
        ):
            raise forms.ValidationError(
                {"cutoff_datetime": "Invalid.  Must be at least 1 day after start date"}
            )

    class Meta:
        model = StockRequest
        fields = "__all__"
        help_text = {"request_identifier": "(read-only)"}  # noqa: RUF012
        widgets = {  # noqa: RUF012
            "request_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
        }



from django.apps import apps as django_apps
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils import timezone

from .choices import TIMEPOINT_STATUS
from .constants import CLOSED_TIMEPOINT, FEEDBACK, OPEN_TIMEPOINT
from .timepoint import TimepointClosed
from .timepoint_collection import TimepointConfigError
from .timepoint_lookup import TimepointLookup
from .utils import get_enable_timepoint_checks


class UnableToCloseTimepoint(Exception):  # noqa: N818
    pass


MODEL_NOT_REGISTERED = (
    "Model not registered. Model '{label_lower}' is not registered "
    "in AppConfig as a timepoint. "
    "See AppConfig for 'edc_timepoint'."
)


class TimepointLookupModelMixin(models.Model):
    """Makes a model lookup the timepoint model instance on `save`
    and check if it is a closed before allowing a create or update.

    Note: the timepoint model uses the TimepointModelMixin, e.g. Appointment
    """

    timepoint_lookup_cls = TimepointLookup

    def save(self, *args, **kwargs):
        if get_enable_timepoint_checks() and getattr(self, "timepoint_lookup_cls", None):
            timepoint_lookup = self.timepoint_lookup_cls()
            if timepoint_lookup.timepoint_model == self._meta.label_lower:
                raise ImproperlyConfigured(
                    f"Timepoint model cannot use TimepointLookupModelMixin. "
                    f"Got {self._meta.label_lower}"
                )
            timepoint_lookup.raise_if_closed(model_obj=self)
        super().save(*args, **kwargs)

    class Meta:
        abstract = True


class TimepointModelMixin(models.Model):
    """Makes a model serve as a marker for a timepoint, e.g. Appointment."""

    enabled_as_timepoint = True

    timepoint_status = models.CharField(
        max_length=15, choices=TIMEPOINT_STATUS, default=OPEN_TIMEPOINT
    )

    timepoint_opened_datetime = models.DateTimeField(
        null=True,
        editable=False,
        help_text="the original calculated model's datetime, updated in the signal",
    )

    timepoint_closed_datetime = models.DateTimeField(null=True, editable=False)

    def save(self, *args, **kwargs):
        if self.enabled_as_timepoint and (
            kwargs.get("update_fields") != ["timepoint_status"]
            and kwargs.get("update_fields")
            != ["timepoint_opened_datetime", "timepoint_status"]
            and kwargs.get("update_fields")
            != ["timepoint_closed_datetime", "timepoint_status"]
        ):
            self.timepoint_open_or_raise()
        super().save(*args, **kwargs)

    def update_timepoint(self):
        """Called by signal"""
        if self.enabled_as_timepoint:
            app_config = django_apps.get_app_config("edc_timepoint")
            if "historical" not in self._meta.label_lower:
                timepoint = app_config.timepoints.get(self._meta.label_lower)
                datetime_value = getattr(self, timepoint.datetime_field)
                if (
                    self.timepoint_opened_datetime is None
                    or self.timepoint_opened_datetime != datetime_value
                ):
                    self.timepoint_opened_datetime = datetime_value
                    self.timepoint_status = OPEN_TIMEPOINT
                    self.save_base(
                        update_fields=[
                            "timepoint_opened_datetime",
                            "timepoint_status",
                        ]
                    )

    def timepoint_open_or_raise(self, timepoint=None):
        if not timepoint:
            app_config = django_apps.get_app_config("edc_timepoint")
            try:
                timepoint = app_config.timepoints.get(self._meta.label_lower)
            except KeyError as e:
                raise TimepointConfigError(
                    MODEL_NOT_REGISTERED.format(label_lower=self._meta.label_lower)
                ) from e
        if getattr(self, timepoint.status_field) != timepoint.closed_status:
            self.timepoint_status = OPEN_TIMEPOINT
            self.timepoint_closed_datetime = None
        elif self.timepoint_status == CLOSED_TIMEPOINT:
            raise TimepointClosed(
                f"This '{self._meta.verbose_name}' instance is closed "
                f"for data entry. See Timepoint."
            )
        return True

    def timepoint_close_timepoint(self):
        """Closes a timepoint.

        Updates the timepoint specific fields when the status field
        changes to closed.
        """
        app_config = django_apps.get_app_config("edc_timepoint")
        timepoint = app_config.timepoints.get(self._meta.label_lower)
        status = getattr(self, timepoint.status_field)
        if status == timepoint.closed_status:
            self.timepoint_status = CLOSED_TIMEPOINT
            self.timepoint_closed_datetime = timezone.now()
            self.save(update_fields=["timepoint_status"])
        else:
            raise UnableToCloseTimepoint(
                f"Unable to close timepoint. Got {self._meta.label_lower}."
                f"{timepoint.status_field} != {timepoint.closed_status}. "
                f"Got '{status}'."
            )

    def timepoint_open_timepoint(self):
        """Re-opens a timepoint."""
        if self.timepoint_status == CLOSED_TIMEPOINT:
            self.timepoint_status = OPEN_TIMEPOINT
            self.timepoint_closed_datetime = None
            self.save(update_fields=["timepoint_closed_datetime", "timepoint_status"])

    def timepoint(self):
        """Formats and returns the status for the change_list."""
        if self.timepoint_status == OPEN_TIMEPOINT:
            return '<span style="color:green;">Open</span>'
        if self.timepoint_status == CLOSED_TIMEPOINT:
            return '<span style="color:red;">Closed</span>'
        if self.timepoint_status == FEEDBACK:
            return '<span style="color:orange;">Feedback</span>'
        return None

    class Meta:
        abstract = True



# Generated by Django 3.0.6 on 2020-07-29 19:40

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("edc_action_item", "0025_auto_20200528_0520"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="actionitem",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "get_latest_by": "modified",
                "ordering": ("-modified", "-created"),
                "verbose_name": "Action Item",
                "verbose_name_plural": "Action Items",
            },
        ),
        migrations.AlterModelOptions(
            name="actiontype",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "ordering": ["name"],
            },
        ),
    ]



from typing import Any

from .exceptions import RegisterRandomizerError
from .randomizer import Randomizer
from .site_randomizers import site_randomizers

__all__ = ["register"]


def register(site=None, **kwargs) -> Any:
    """Registers a randomizer class."""
    site = site or site_randomizers

    def _register_randomizer_cls_wrapper(randomizer_cls: Any) -> Any:
        if not issubclass(randomizer_cls, (Randomizer,)):
            raise RegisterRandomizerError(
                f"Wrapped class must a Randomizer class. Got {randomizer_cls}"
            )
        site.register(randomizer_cls)
        return randomizer_cls

    return _register_randomizer_cls_wrapper



from django import forms

from ...exceptions import OffstudyError
from ...utils import raise_if_offstudy


class OffstudyNonCrfModelFormMixin:
    """ModelForm mixin for non-CRF modelforms / PRNs."""

    def clean(self):
        cleaned_data = super().clean()
        self.raise_if_offstudy_by_report_datetime()
        return cleaned_data

    def raise_if_offstudy_by_report_datetime(self) -> None:
        if self.get_subject_identifier() and self.report_datetime:
            try:
                raise_if_offstudy(
                    source_obj=self.instance,
                    subject_identifier=self.get_subject_identifier(),
                    report_datetime=self.report_datetime,
                )
            except OffstudyError as e:
                raise forms.ValidationError(e)



from clinicedc_constants import NO, YES
from django.contrib import admin
from django.contrib.admin import SimpleListFilter
from django.core.exceptions import ObjectDoesNotExist
from django.template.loader import render_to_string
from django.urls import reverse
from django_audit_fields import audit_fieldset_tuple
from edc_model_admin.history import SimpleHistoryAdmin
from edc_utils.date import to_local
from rangefilter.filters import DateRangeFilterBuilder

from ...admin_site import edc_pharmacy_admin
from ...models import StockTransferItem
from ..model_admin_mixin import ModelAdminMixin


class ConfirmedAtLocationFilter(SimpleListFilter):
    title = "Confirmed at location"
    parameter_name = "confirmed_at_location"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return (YES, YES), (NO, NO)

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            opts = dict(
                stock__from_stock__isnull=False,
                stock__confirmation__isnull=False,
                stock__allocation__isnull=False,
            )
            if self.value() == YES:
                qs = queryset.filter(
                    confirmationatlocationitem__isnull=False,
                    **opts,
                )
            elif self.value() == NO:
                qs = queryset.filter(
                    confirmationatlocationitem__isnull=True,
                    **opts,
                )
        return qs


@admin.register(StockTransferItem, site=edc_pharmacy_admin)
class StockTransferItemAdmin(ModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Stock Transfer Item"
    change_form_title = "Pharmacy: Stock Transfer Items"
    change_list_note = "D=Dispensed"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20

    autocomplete_fields = ("stock",)

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "transfer_item_identifier",
                    "transfer_item_datetime",
                    "stock_transfer",
                    "stock",
                )
            },
        ),
        audit_fieldset_tuple,
    )
    list_display = (
        "identifier",
        "stock_transfer_changelist",
        "transfer_item_date",
        "stock_changelist",
        "allocation_changelist",
        "confirmation_at_location_item_changelist",
        "dispensed",
        "formatted_from_location",
        "formatted_to_location",
    )

    list_filter = (
        "stock_transfer__to_location",
        ("transfer_item_datetime", DateRangeFilterBuilder()),
        ConfirmedAtLocationFilter,
        "stock__dispensed",
    )

    search_fields = (
        "id",
        "transfer_item_identifier",
        "stock_transfer__id",
        "stock__code",
        "stock__allocation__registered_subject__subject_identifier",
    )

    readonly_fields = (
        "transfer_item_identifier",
        "transfer_item_datetime",
        "stock",
        "stock_transfer",
    )

    @admin.display(description="TRANSFER ITEM #", ordering="transfer_item_identifier")
    def identifier(self, obj):
        return obj.transfer_item_identifier

    @admin.display(description="D", ordering="stock__dispensed", boolean=True)
    def dispensed(self, obj):
        return obj.stock.dispensed

    @admin.display(description="From", ordering="stock__location")
    def formatted_from_location(self, obj):
        return obj.stock_transfer.from_location.display_name

    @admin.display(description="To", ordering="stock__location")
    def formatted_to_location(self, obj):
        if obj:
            is_current_location = obj.stock_transfer.to_location == obj.stock.location
            context = dict(
                location=obj.stock_transfer.to_location.display_name,
                is_current_location=is_current_location,
            )
            return render_to_string("edc_pharmacy/stock/location.html", context=context)
        return None

    @admin.display(description="Loc", ordering="stock__location")
    def formatted_location(self, obj):
        try:
            obj.confirmationatlocationitem  # noqa: B018
        except ObjectDoesNotExist:
            return f"{obj.stock.location.display_name} > {obj.stock_transfer.to_location.display_name}"
        else:
            return obj.stock.location.display_name

    @admin.display(description="Transfer date", ordering="transfer_item_datetime")
    def transfer_item_date(self, obj):
        return to_local(obj.transfer_item_datetime).date()

    @admin.display(description="Stock #", ordering="stock__code")
    def stock_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stock_changelist")
        url = f"{url}?q={obj.stock.code}"
        context = dict(url=url, label=obj.stock.code, title="Go to stock")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(
        description="Allocation",
        ordering="stock__allocation__registered_subject__subject_identifier",
    )
    def allocation_changelist(self, obj):
        subject_identifier = obj.stock.allocation.registered_subject.subject_identifier
        url = reverse("edc_pharmacy_admin:edc_pharmacy_allocation_changelist")
        url = f"{url}?q={subject_identifier}"
        context = dict(url=url, label=subject_identifier, title="Go to allocation")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(description="Transfer #", ordering="stock_transfer__transfer_identifier")
    def stock_transfer_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stocktransfer_changelist")
        url = f"{url}?q={obj.stock_transfer.id}"
        context = dict(
            url=url,
            label=obj.stock_transfer.transfer_identifier,
            title="Go to stock transfer",
        )
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(
        description="Confirmation #",
        ordering="stock__confirmationatlocationitem__transfer_confirmation_item_identifier",
    )
    def confirmation_at_location_item_changelist(self, obj):
        try:
            transfer_confirmation_item = obj.confirmationatlocationitem
        except ObjectDoesNotExist:
            url = reverse("edc_pharmacy:confirm_at_location_url")
            context = dict(
                url=url,
                label="Pending",
                title="Go to stock transfer site confirmation",
            )
        else:
            url = reverse(
                "edc_pharmacy_admin:edc_pharmacy_confirmationatlocationitem_changelist"
            )
            url = f"{url}?q={transfer_confirmation_item.pk}"
            context = dict(
                url=url,
                label=transfer_confirmation_item.transfer_confirmation_item_identifier,
                title="Go to stock transfer confirmation item",
            )
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)



from django.core.exceptions import ObjectDoesNotExist
from django.db import models, transaction

from edc_consent import site_consents
from edc_consent.exceptions import ConsentDefinitionModelError
from edc_sites import site_sites


class ConsentVersionModelMixin(models.Model):
    """A model mixin that adds version to a consent.

    Requires at least `NonUniqueSubjectIdentifierModelMixin` (or
    `UniqueSubjectIdentifierModelMixin`), `SiteModelMixin` and
    the field `consent_datetime`.
    """

    version = models.CharField(
        verbose_name="Consent version",
        max_length=10,
        help_text="See 'consent definition' for consent versions by period.",
        editable=False,
    )

    update_versions = models.BooleanField(default=False)

    consent_definition_name = models.CharField(
        verbose_name="Consent definition", max_length=50, default="", editable=False
    )

    def __str__(self):
        return f"{self.get_subject_identifier()} v{self.version}"

    def save(self, *args, **kwargs):
        cdef = self.consent_definition
        self.version = cdef.version
        self.consent_definition_name = cdef.name
        if not self.id and self.subject_identifier:
            try:
                with transaction.atomic():
                    previous_consent = cdef.get_previous_consent(
                        subject_identifier=self.subject_identifier, exclude_id=self.id
                    )
            except ObjectDoesNotExist:
                pass
            else:
                self.first_name = previous_consent.first_name
                self.dob = previous_consent.dob
                self.initials = previous_consent.initials
                self.identity = previous_consent.identity
                self.confirm_identity = previous_consent.confirm_identity
        super().save(*args, **kwargs)

    @property
    def consent_definition(self):
        """Allow the consent to save as long as there is a
        consent definition for this report_date and site.
        """
        site = self.site
        if not self.id and not site:
            site = site_sites.get_current_site_obj()
        cdef = site_consents.get_consent_definition(
            model=self._meta.label_lower,
            report_datetime=self.consent_datetime,
            site=site_sites.get(site.id),
        )
        if self._meta.label_lower != cdef.model:
            raise ConsentDefinitionModelError(
                f"Incorrect model for consent_definition. This model cannot be used "
                f"to 'create' consent version '{cdef.version}'. Expected "
                f"'{cdef.model}'. Got '{self._meta.label_lower}'."
            )
        if cdef.updates and self._meta.label_lower != cdef.updates.updated_by.model:
            raise ConsentDefinitionModelError(
                f"Incorrect model to update a consent. This model cannot be used "
                f"to 'update' consent version '{cdef.version}'. Expected "
                f"'{cdef.updates.updated_by.model}'. Got '{self._meta.label_lower}'."
            )
        return cdef

    class Meta:
        abstract = True



from clinicedc_constants import MILLIMOLES_PER_LITER
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models

from ..model_mixin_factories import reportable_result_model_mixin_factory


class CholModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="chol",
        verbose_name="Total Cholesterol",
        decimal_places=2,
        max_digits=8,
        units_choices=((MILLIMOLES_PER_LITER, MILLIMOLES_PER_LITER),),
        validators=[MinValueValidator(0.00), MaxValueValidator(999.00)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class HdlModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="hdl",
        decimal_places=2,
        max_digits=8,
        units_choices=((MILLIMOLES_PER_LITER, MILLIMOLES_PER_LITER),),
        validators=[MinValueValidator(0.00), MaxValueValidator(999.00)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class LdlModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="ldl",
        decimal_places=2,
        max_digits=8,
        units_choices=((MILLIMOLES_PER_LITER, MILLIMOLES_PER_LITER),),
        validators=[MinValueValidator(0.00), MaxValueValidator(999.00)],
    ),
    models.Model,
):
    class Meta:
        abstract = True


class TrigModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="trig",
        decimal_places=2,
        max_digits=8,
        units_choices=((MILLIMOLES_PER_LITER, MILLIMOLES_PER_LITER),),
        validators=[MinValueValidator(0.00), MaxValueValidator(999.00)],
        verbose_name="Triglycerides",
    ),
    models.Model,
):
    class Meta:
        abstract = True
